from fsgs.plugins.pluginmanager import PluginManager

"""
Debug script used to dump information about detected plugins
"""


def app_main():
    plugin_manager = PluginManager.instance()
    plugins = plugin_manager.plugins()
    provides = plugin_manager.provides()

    for plugin in plugins:
        print("# {} ({})".format(plugin.name, type(plugin).__name__))
        print(plugin.path)
        print("")

    print("# Provides:")
    for name in sorted(provides):
        plugin = provides[name]
        print("# {} ({})".format(name, plugin.name))
        # print("", plugin.provides()[name])

    print("")



import json
from urllib.parse import parse_qsl

from fsgs.http.requestparams import RequestParams


class Request:
    def __init__(self, environ):
        self.environ = environ
        self.session_id = ""
        self.fsgs = None
        self.params = RequestParams()

        ct = environ.get("CONTENT_TYPE", "").lower()
        cl = environ.get("CONTENT_LENGTH", "")
        if cl:
            cl = int(cl)
        else:
            cl = 0
        print(ct)
        if ct in ["application/json", "application/json;charset=utf-8"]:
            input = environ["wsgi.input"]
            print("reading...")
            body = input.read(cl)
            print("reading done")
            print(body)
            body_object = json.loads(body.decode("UTF-8"))
            print(body_object)
            if isinstance(body_object, dict):
                for key, value in body_object.items():
                    setattr(self.params, key, value)

        query_string = environ.get("QUERY_STRING", "")
        items = parse_qsl(
            query_string, keep_blank_values=False, strict_parsing=False
        )
        for item in items:
            setattr(self.params, item[0], item[1])



import fsui.qt
from .widget_mixin import WidgetMixin


class TextArea(fsui.qt.QTextEdit, WidgetMixin):
    changed = fsui.qt.Signal()

    def __init__(
        self, parent, text="", read_only=False, font_family=None, border=True
    ):
        fsui.qt.QTextEdit.__init__(self, "", parent.get_container())
        # Widget.__init__(self, parent)
        self.init_widget(parent)
        if not border:
            self.setFrameStyle(fsui.qt.QFrame.Shape.NoFrame)
        self.setReadOnly(read_only)
        if font_family:
            print("FIXME: not respecting font_family yet")
            font = fsui.qt.QFont("Courier")
            self.setFont(font)
        if text:
            self.append_text(text)
        self.textChanged.connect(self.__text_changed)

    def get_text(self):
        return self.toPlainText()

    def set_text(self, text):
        self.setPlainText(text.replace("\n", "\r\n"))

    def append_text(self, text, color=None):
        # text = text.replace("\n", "\r\n")
        # print("Appending text:", repr(text))
        # self.moveCursor(QTextCursor.MoveOperation.End)
        # self.insertPlainText(text.strip())
        if color is not None:
            self.setTextColor(fsui.qt.QColor(*color))
        # self.appendPlainText(text.strip())
        self.append(text.strip())
        self.moveCursor(fsui.qt.QTextCursor.MoveOperation.End)

    def scroll_to_start(self):
        self.moveCursor(fsui.qt.QTextCursor.MoveOperation.Start)

    def __text_changed(self):
        self.changed.emit()



from __future__ import absolute_import
from __future__ import print_function

from .MetaInfo import MetaInfo
from .RootMetaInfo import RootMetaInfo
from .ProtectFlags import ProtectFlags
from .TimeStamp import TimeStamp
from amitools.fs.block.BootBlock import BootBlock
from . import DosType
from .FSString import FSString

class MetaDB:
  def __init__(self):
    self.metas = {}
    self.vol_name = None
    self.vol_meta = None
    self.dos_type = DosType.DOS0
  
  def set_root_meta_info(self, meta):
    self.vol_meta = meta
  
  def get_root_meta_info(self):
    return self.vol_meta
  
  def set_volume_name(self, name):
    if type(name) != unicode:
      raise ValueError("set_volume_name must be unicode")
    self.vol_name = name
  
  def get_volume_name(self):
    return self.vol_name
  
  def set_dos_type(self, dos_type):
    self.dos_type = dos_type
  
  def get_dos_type(self):
    return self.dos_type
  
  def set_meta_info(self, path, meta_info):
    if type(path) != unicode:
      raise ValueError("set_meta_info: path must be unicode")
    self.metas[path] = meta_info
  
  def get_meta_info(self, path):
    if self.metas.has_key(path):
      return self.metas[path]
    else:
      return None
  
  def dump(self):
    print(self.vol_name, self.vol_meta, self.dos_type)
    for m in self.metas:
      print(m)
  
  # ----- load -----
    
  def load(self, file_path):
    self.metas = {}
    f = open(file_path, "r")
    first = True
    for line in f:
      if first:
        self.load_header(line)
        first = False
      else:
        self.load_entry(line)
    f.close()
    
  def load_header(self, line):
    pos = line.find(':')
    if pos == -1:
      raise IOError("Invalid xdfmeta header! (no colon in line)")
    # first extract volume name
    vol_name = line[:pos]
    self.vol_name = vol_name.decode("UTF-8")
    line = line[pos+1:]
    # now get parameters
    comp = line.split(',')
    if len(comp) != 4:
      raise IOError("Invalid xdfmeta header! (wrong number of parameters found)")
    # first dos type  
    dos_type_str = comp[0]
    if len(dos_type_str) != 4:
      raise IOError("Invalid xdfmeta dostype string")
    num = ord(dos_type_str[3]) - ord('0')
    if num < 0 or num > 5:
      raise IOError("Invalid xdfmeta dostype number")
    self.dos_type = DosType.DOS0 + num
    # then time stamps
    create_ts = TimeStamp()
    ok1 = create_ts.parse(comp[1])
    disk_ts = TimeStamp()
    ok2 = disk_ts.parse(comp[2])
    mod_ts = TimeStamp()
    ok3 = mod_ts.parse(comp[3])
    if not ok1 or not ok2 or not ok3:
      raise IOError("Invalid xdfmeta header! (invalid timestamp found)")
    self.vol_meta = RootMetaInfo(create_ts, disk_ts, mod_ts)
    
  def load_entry(self, line):
    line = line.strip()
    # path
    pos = line.find(':')
    if pos == -1:
      raise IOError("Invalid xdfmeta file! (no colon in line)")
    path = line[:pos].decode("UTF-8")
    # prot
    line = line[pos+1:]
    pos = line.find(',')
    if pos == -1:
      raise IOError("Invalid xdfmeta file! (no first comma)")
    prot_str = line[:pos]
    prot = ProtectFlags()
    prot.parse(prot_str)
    # time
    line = line[pos+1:]
    pos = line.find(',')
    if pos == -1:
      raise IOError("Invalid xdfmeta file! (no second comma)")
    time_str = line[:pos]
    time = TimeStamp()
    time.parse(time_str)
    # comment
    comment = FSString(line[pos+1:].decode("UTF-8"))
    # meta info
    mi = MetaInfo(protect_flags=prot, mod_ts=time, comment=comment)
    self.set_meta_info(path, mi)
  
  # ----- save -----
  
  def save(self, file_path):
    f = open(file_path, "w")
    # header
    mi = self.vol_meta
    num = self.dos_type - DosType.DOS0 + ord('0')
    dos_type_str = "DOS%c" % num
    vol_name = self.vol_name.encode("UTF-8")
    line = "%s:%s,%s,%s,%s\n" % (vol_name, dos_type_str, mi.get_create_ts(), mi.get_disk_ts(), mi.get_mod_ts())
    f.write(line)
    # entries
    for path in sorted(self.metas):
      meta_info = self.metas[path]
      protect = meta_info.get_protect_short_str()
      mod_time = meta_info.get_mod_time_str()
      comment = meta_info.get_comment_unicode_str().encode("UTF-8")
      path_name = path.encode("UTF-8")
      line = "%s:%s,%s,%s\n" % (path_name, protect, mod_time, comment)
      f.write(line)
    f.close()



import io
import os
import sys
from configparser import ConfigParser

import fsbc.application
from fsgs.application import ApplicationMixin
from fsgs.context import fsgs
from fsgs.Database import Database
from fsgs.input.enumeratehelper import EnumerateHelper
from fsgs.platform import PlatformHandler
from fsgs.util.gamenameutil import GameNameUtil


class Application(ApplicationMixin, fsbc.application.Application):
    pass


def main():
    Application("fs-uae-game-system")

    if "--unsupported" in sys.argv:
        if "--http-server" in sys.argv:
            from fsgs.http.server import http_server_main

            return http_server_main()

    if len(sys.argv) < 3:
        print("")
        print("usage: fsgs run <game>")
        print("")
        print("game:")
        print(" - search term(s) identifying a single game")
        print(" - path to a .fsgs file")
        print(" - path to a recognized cartridge ROM or disk file format")
        print("")
        sys.exit(1)
    assert sys.argv[1] == "run"
    game_arg = " ".join(sys.argv[2:])
    print(game_arg)
    if os.path.exists(game_arg):
        load_file(game_arg)
    else:
        search = game_arg.lower()
        database = Database.instance()
        # cursor.execute("SELECT id FROM game WHERE name like")
        terms = GameNameUtil.extract_search_terms(search)
        found_games = database.find_games_new(" ".join(terms))
        games = []
        for game in found_games:
            print(list(game))
            if game[0]:
                # only process entries with a game uuid
                games.append(game)
        game_uuid = None
        if len(games) == 0:
            print("no games found")
            sys.exit(2)
        if len(games) > 1:
            matches = 0
            for row in games:
                if row[1].lower() == search:
                    if game_uuid is None:
                        game_uuid = row[0]
                        matches += 1
            if matches != 1:
                print("")
                print("More than one game matches:")
                print("")
                for row in games:
                    print("    {0} ({1})".format(row[1], row[2]))
                    print("        {0}".format(row[0]))
                print("")
                sys.exit(3)
        game_uuid = games[0][0]
        assert game_uuid
        variant_uuid = find_preferred_variant(game_uuid)
        load_game_variant(variant_uuid)
    fsgs.run_game()


def find_preferred_variant(game_uuid):
    return fsgs.find_preferred_game_variant(game_uuid)


def load_game_variant(variant_uuid):
    return fsgs.load_game_variant(variant_uuid)


def load_file(path):
    config = {}
    name, ext = os.path.splitext(path)
    if ext in [".fs-uae", ".fsgs"]:
        return load_config_file(path)
    elif ext == ".st":
        config["platform"] = "atari-st"
        config["floppy_drive_0"] = path
    elif ext in [".adf", ".dms", ".ipf"]:
        config["platform"] = "amiga"
        config["floppy_drive_0"] = path
    elif ext in [".tap"]:
        config["platform"] = "commodore-64"
        config["tape_drive"] = path
    if config:
        load_config(config)


def load_config_file(fsgs_file):
    cp = ConfigParser()
    with io.open(fsgs_file, "r", encoding="UTF-8") as f:
        cp.read_file(f)
    config = {}
    if cp.has_section("fsgs"):
        for key in cp.options("fsgs"):
            value = cp.get("fsgs", key)
            config[key.lower().replace("-", "_")] = value
    load_config(config)


def load_config(config):
    fsgs.config.load(config)
    fsgs.game.platform.id = config["platform"]
    fsgs.game.uuid = "7bc9ae8b-e454-4108-87fe-6aac09cfb1e9"
    fsgs.game.name = "Default Game"
    fsgs.game.variant.uuid = "973d787f-2cc4-4d8d-b0c1-1bd911ef407a"
    fsgs.game.variant.name = "Default Variant"



import platform as _platform
import sys

windows = sys.platform == "win32"
linux = sys.platform.startswith("linux")
macosx = sys.platform == "darwin"

if windows:
    platform = "windows"
elif linux:
    platform = "linux"
elif macosx:
    platform = "macos"
else:
    platform = "unknown"


class System:
    windows = windows
    linux = linux
    macos = macosx
    platform = platform

    x86_64 = False
    if _platform.machine().lower() in [
        "x86_64",
        "x86-64",
        "amd64",
        "i386",
        "i486",
        "i586",
        "i686",
    ]:
        if _platform.architecture()[0] == "64bit":
            x86_64 = True
        # else:
        #     x86 = True



from fsbc.application import app
from fsgs.ogd.base import SynchronizerBase
from fsgs.res import gettext


class MetaSynchronizer(SynchronizerBase):
    def __init__(self, *args, **kwargs):
        SynchronizerBase.__init__(self, *args, **kwargs)

    def synchronize(self):
        if self.stop_check():
            return
        if not app.settings["database_auth"]:
            # not logged in
            return
        self.set_status(gettext("Fetching synchronization information..."))
        self.context.meta = self.fetch_json("/api/sync")



from fsbc.system import linux, macosx, windows


class Device(object):
    TYPE_KEYBOARD = "keyboard"
    TYPE_JOYSTICK = "joystick"
    TYPE_MOUSE = "mouse"

    def __init__(self):
        self.id = ""
        self.name = ""
        # The unmodified sdl_name of the device (non-stripped)
        self.sdl_name = ""
        self.type = ""
        # self.port = None
        # self.cmp_id = create_cmp_id(id)

        self.index = 0
        self.balls = 0
        self.hats = 0
        self.axes = 0
        self.buttons = 0
        # self.event_id = ""

    def __lt__(self, other):
        if self.index < other.index:
            return True
        return self.name < other.name

    # Returns the unmodified sdl_name of the device (non-stripped)
    # @property
    # def sdl_name(self):
    #     return self.sdl_name

    def is_keyboard(self):
        # print("id:", self.id)
        # print("name:", self.name)
        # print("type:", self.type)
        return self.type == self.TYPE_KEYBOARD

    def get_config_name(self):
        name = self.name.rsplit("#", 1)[0]
        name_lower = name.lower()
        name = ""
        for c in name_lower:
            if c in "abcdefghijklmnopqrstuvwxyz0123456789":
                name += c
            else:
                if not name.endswith("_"):
                    name += "_"
        name = name.strip("_")

        if windows:
            host_platform = "windows"
        elif macosx:
            host_platform = "macos"
        elif linux:
            host_platform = "linux"
        else:
            host_platform = "other"

        config_name = "{0}_{1}_{2}_{3}_{4}_{5}".format(
            name, self.buttons, self.axes, self.hats, self.balls, host_platform
        )
        return config_name

    def configure(self, system, multiple=True):
        # name = self.name.rsplit("#", 1)[0]
        name = self.name
        from fsgs.input.inputdevice import InputDevice

        try:
            # device id must end with #something (really a device number,
            # but can be anything
            if "#" in name:
                name_with_hash = name
            else:
                name_with_hash = name + " #DUMMY"
            device = InputDevice(
                system,
                name_with_hash,
                [],
                version=2,
                axes=self.axes,
                hats=self.hats,
                buttons=self.buttons,
                balls=self.balls,
                multiple=multiple,
            )
            config = device.get_config()
        except Exception as e:
            print(
                "error initializing device {0} for {1}".format(
                    self.name, system
                )
            )
            print(repr(e))
            # return {}
            raise e
        # config_inv = []
        for key, val in list(config.items()):
            val = val.upper()
            config[key] = val
            config[val] = key
        return config

    def __repr__(self):
        return "<Device '{0}'>".format(self.id)



import os
import threading
import time
from functools import lru_cache
from uuid import uuid4
from zipfile import ZipFile

import requests

from arcade.glui.texturemanager import TextureManager
from arcade.resources import logger
from fsgs.FSGSDirectories import FSGSDirectories
from fsgs.network import openretro_url_prefix
from fsui.qt import QImage

error_set = set()


@lru_cache()
def get_cache_zip_for_sha1(sha1):
    zip_path = os.path.join(FSGSDirectories.images_dir(), sha1[:2] + ".zip")
    try:
        return ZipFile(zip_path, "r")
    except Exception:
        return None


def get_file_for_sha1_cached(sha1, size_arg, cache_ext):
    cache_zip = get_cache_zip_for_sha1(sha1)
    if cache_zip is not None:
        try:
            return cache_zip.open("{}/{}{}".format(sha1[:2], sha1, cache_ext))
        except KeyError:
            pass
    cache_dir = FSGSDirectories.images_dir_for_sha1(sha1)
    cache_file = os.path.join(cache_dir, sha1 + cache_ext)
    if os.path.exists(cache_file):
        # An old bug made it possible for 0-byte files to exist, so
        # we check for that here...
        if os.path.getsize(cache_file) > 0:
            return cache_file

    url = "{}/image/{}{}".format(openretro_url_prefix(), sha1, size_arg)
    print("[IMAGES]", url)

    r = requests.get(url, stream=True)
    try:
        r.raise_for_status()
        cache_file_partial = "{}.{}.partial".format(
            cache_file, str(uuid4())[:8]
        )
        if not os.path.exists(os.path.dirname(cache_file_partial)):
            os.makedirs(os.path.dirname(cache_file_partial))
        with open(cache_file_partial, "wb") as f:
            for chunk in r.iter_content(chunk_size=65536):
                f.write(chunk)
    finally:
        r.close()
    os.rename(cache_file_partial, cache_file)
    return cache_file


def get_file_for_sha1(sha1):
    sha1, size_arg = sha1.split("?")
    if size_arg == "s=1x":
        cache_ext = "_1x.png"
    elif size_arg == "s=512&f=jpg":
        cache_ext = "_512.jpg"
    elif size_arg == "w=480&h=640&t=cc&f=jpg":
        cache_ext = "_480x640_cc.jpg"
    else:
        raise Exception("unrecognized size")
    size_arg = "?" + size_arg
    return get_file_for_sha1_cached(sha1, size_arg, cache_ext)


def load_image(relative_path):
    path = ""
    try:
        if relative_path.startswith("sha1:"):
            sha1 = relative_path[5:]
            path = get_file_for_sha1(sha1)
        else:
            path = relative_path

        if relative_path in error_set:
            return None, (0, 0)

        if hasattr(path, "read"):
            im = QImage()
            im.loadFromData(path.read())
        else:
            if not os.path.exists(path):
                return None, (0, 0)
            im = QImage(path)
        if im.format() != QImage.Format.Format_ARGB32:
            im = im.convertToFormat(QImage.Format.Format_ARGB32)
        bits = im.bits()
        try:
            pixels = bits.tobytes()
        except AttributeError:
            bits.setsize(im.sizeInBytes())
            pixels = bytes(bits)
        return pixels, (im.width(), im.height())

    except Exception as e:
        print(
            "[IMAGES] Error loading", repr(relative_path), repr(path), repr(e)
        )
        error_set.add(relative_path)
        return None, (0, 0)


class ImageLoader(object):
    _instance = None

    @classmethod
    def get(cls):
        if not cls._instance:
            cls._instance = cls()
        return cls._instance

    def __init__(self):
        self._stop_flag = False

    def start(self):
        self._stop_flag = False
        threading.Thread(
            target=self.image_loader_thread, name="GameCenterImageLoaderThread"
        ).start()
        pass

    def image_loader_thread(self):
        logger.debug("[IMAGES] Image loader started")
        tm = TextureManager().get()
        while not self._stop_flag:
            time.sleep(0.01)

            load_ip = None
            with tm.lock:
                for ip in tm.image_list:
                    if tm.image_dict[ip][0] is False:
                        load_ip = ip
                        break
            if load_ip:
                pixels, size = load_image(load_ip)
                tm.set_image(load_ip, pixels, size)

    def stop(self):
        print("[IMAGES] ImageLoader.stop()")
        self._stop_flag = True



import threading
import traceback

import requests

import launcher.version
from fsbc.application import Application
from fsbc.system import linux, macosx, windows
from fsbc.util import compare_versions, unused
from fstd.desktop import open_url_in_browser

from .launcher_settings import LauncherSettings
from .launcher_signal import LauncherSignal


class UpdateManager:
    @classmethod
    def run_update_check(cls):
        threading.Thread(target=cls.update_thread_function).start()

    @classmethod
    def update_thread_function(cls):
        try:
            cls._update_thread_function()
        except Exception:
            traceback.print_exc()

    @classmethod
    def series(cls):
        if "dev" in Application.instance().version:
            return "devel"
        elif "beta" in Application.instance().version:
            return "beta"
        return "stable"

    @classmethod
    def _update_thread_function(cls):
        if windows:
            platform = "windows"
        elif macosx:
            platform = "macosx"
        elif linux:
            platform = "linux"
        else:
            platform = "other"
        url = "https://fs-uae.net/{0}/latest-{1}".format(
            cls.series(), platform
        )
        r = requests.get(url)
        r.raise_for_status()
        version_str = r.text.strip()
        print("latest version available: ", version_str)
        print("current version: ", launcher.version.VERSION)
        result = compare_versions(version_str, launcher.version.VERSION)
        print("update check result: ", result)
        if result > 0 and version_str != "9.9.9":
            web_url = "https://fs-uae.net/{0}/download/".format(cls.series())
            LauncherSignal.broadcast("update_available", version_str, web_url)
            # FIXME: Thread safety...
            # FIXME: Use above signal instead
            LauncherSettings.set("__update_available", version_str)

    @classmethod
    def start_update(cls, version_str):
        unused(version_str)
        web_url = "https://fs-uae.net/{0}/download/".format(cls.series())
        open_url_in_browser(web_url)



# Copyright (c) 2008 Duncan Fordyce
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# The above copyright notice and this permission notice shall be included in
#  all copies or substantial portions of the Software.
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging

from oyoyo.ircevents import *


def parse_raw_irc_command(element):
    """
    This function parses a raw irc command and returns a tuple
    of (prefix, command, args).
    The following is a psuedo BNF of the input text:

    <message>  ::= [':' <prefix> <SPACE> ] <command> <params> <crlf>
    <prefix>   ::= <servername> | <nick> [ '!' <user> ] [ '@' <host> ]
    <command>  ::= <letter> { <letter> } | <number> <number> <number>
    <SPACE>    ::= ' ' { ' ' }
    <params>   ::= <SPACE> [ ':' <trailing> | <middle> <params> ]

    <middle>   ::= <Any *non-empty* sequence of octets not including SPACE
                   or NUL or CR or LF, the first of which may not be ':'>
    <trailing> ::= <Any, possibly *empty*, sequence of octets not including
                     NUL or CR or LF>

    <crlf>     ::= CR LF
    """
    print(element)
    element = element.decode("UTF-8", errors="replace")
    parts = element.strip().split(" ")
    if parts[0].startswith(":"):
        prefix = parts[0][1:]
        command = parts[1]
        args = parts[2:]
    else:
        prefix = None
        command = parts[0]
        args = parts[1:]

    if command.isdigit():
        try:
            command = numeric_events[command]
        except KeyError:
            logging.warning('unknown numeric event %s' % command)
    command = command.lower()

    if args[0].startswith(":"):
        args = [" ".join(args)[1:]]
    else:
        for idx, arg in enumerate(args):
            if arg.startswith(":"):
                args = args[:idx] + [" ".join(args[idx:])[1:]]
                break

    return prefix, command, args


def parse_nick(name):
    """ parse a nickname and return a tuple of (nick, mode, user, host)

    <nick> [ '!' [<mode> = ] <user> ] [ '@' <host> ]
    """

    try:
        nick, rest = name.split('!')
    except ValueError:
        return name, None, None, None
    try:
        mode, rest = rest.split('=')
    except ValueError:
        mode, rest = None, rest
    try:
        user, host = rest.split('@')
    except ValueError:
        return name, mode, rest, None

    return name, mode, user, host



"""Scan an ADF image an visit all files"""

import StringIO

from amitools.fs.blkdev.BlkDevFactory import BlkDevFactory
from amitools.fs.ADFSVolume import ADFSVolume

class ADFSScanner:

  def __init__(self):
    self.factory = BlkDevFactory()

  def can_handle(self, scan_file):
    base_name = scan_file.get_basename().lower()
    for ext in self.factory.valid_extensions:
      if base_name.endswith(ext):
        return True
    return False

  def handle(self, scan_file, scanner):
    if scan_file.is_seekable():
      sf = scan_file
    else:
      sf = scanner.promote_scan_file(scan_file, seekable=True)
    # create blkdev
    blkdev = self.factory.open(sf.get_local_path(), fobj=sf.get_fobj())
    # create volume
    volume = ADFSVolume(blkdev)
    volume.open()
    # scan volume
    node = volume.get_root_dir()
    ok = self._scan_node(sf, scanner, node)
    # done
    volume.close()
    blkdev.close()
    return ok

  def _scan_node(self, scan_file, scanner, node):
    if node.is_dir():
      # recurse into dir
      entries = node.get_entries()
      for e in entries:
        ok = self._scan_node(scan_file, scanner, e)
        if not ok:
          return False
      return True
    elif node.is_file():
      # read file in ram fobj
      data = node.get_file_data()
      node.flush()
      size = len(data)
      path = node.get_node_path_name().get_unicode()
      fobj = StringIO.StringIO(data)
      sf = scan_file.create_sub_path(path, fobj, size, True, False)
      ok = scanner.scan_obj(sf)
      sf.close()
      return True

# mini test
if __name__ == '__main__':
  import sys
  from FileScanner import FileScanner

  ifs = ['*.txt']
  def handler(scan_file):
    print(scan_file)
    return True
  def skip_handler(scan_file):
    print("SKIP:", scan_file)
    return True
  def error_handler(scan_file, error):
    print("FAILED:", scan_file, error)
    raise error
  scanners = [ADFSScanner()]
  fs = FileScanner(handler, ignore_filters=ifs, error_handler=error_handler,
                   scanners=scanners, skip_handler=skip_handler)
  for a in sys.argv[1:]:
    fs.scan(a)



import os
import stat
import amitools.util.BlkDevTools as BlkDevTools
import zlib
import io

class ImageFile:
  def __init__(self, file_name, read_only=False, block_bytes=512, fobj=None):
    self.file_name = file_name
    self.read_only = read_only
    self.block_bytes = block_bytes
    self.fobj = fobj
    self.fh = None
    self.size = 0
    self.num_blocks = 0

  def open(self):
    # file obj?
    if self.fobj is not None:
      self.fh = self.fobj
      # get size via seek
      self.fobj.seek(0,2) # end of file
      self.size = self.fobj.tell()
      self.fobj.seek(0,0) # return to begin
      self.num_blocks = self.size / self.block_bytes
    # file name given
    else:
      # is readable?
      if not os.access(self.file_name, os.R_OK):
        raise IOError("Can't read from image file")
      # is writeable?
      if not os.access(self.file_name, os.W_OK):
        self.read_only = True
      # is it a block/char device?
      st = os.stat(self.file_name)
      mode = st.st_mode
      if stat.S_ISBLK(mode) or stat.S_ISCHR(mode):
        self.size = BlkDevTools.getblkdevsize(self.file_name)
      else:
        # get size and make sure its not empty
        self.size = os.path.getsize(self.file_name)
      if self.size == 0:
        raise IOError("Empty image file detected!")
      self.num_blocks = self.size / self.block_bytes
      # open raw file
      if self.read_only:
        flags = "rb"
      else:
        flags = "r+b"
      self.fh = io.open(self.file_name, flags)

  def read_blk(self, blk_num):
    if blk_num >= self.num_blocks:
      raise IOError("Invalid image file block num: got %d but max is %d" % (blk_num, self.num_blocks))
    off = blk_num * self.block_bytes
    if off != self.fh.tell():
      self.fh.seek(off, os.SEEK_SET)
    num = self.block_bytes
    data = self.fh.read(self.block_bytes)
    return data

  def write_blk(self, blk_num, data):
    if self.read_only:
      raise IOError("Can't write block: image file is read-only")
    if blk_num >= self.num_blocks:
      raise IOError("Invalid image file block num: got %d but max is %d" % (blk_num, self.num_blocks))
    if len(data) != self.block_bytes:
      raise IOError("Invalid block size written: got %d but size is %d" % (len(data), self.block_bytes))
    off = blk_num * self.block_bytes
    if off != self.fh.tell():
      self.fh.seek(off, os.SEEK_SET)
    self.fh.write(data)

  def flush(self):
    self.fh.flush()

  def close(self):
    if self.fh != None:
      self.fh.close()
      self.fh = None

  def create(self, num_blocks):
    if self.read_only:
      raise IOError("Can't create image file in read only mode")
    block = '\0' * self.block_bytes
    if self.fobj is not None:
      for i in xrange(num_blocks):
        self.fobj.write(block)
      self.fobj.seek(0,0)
    else:
      fh = file(self.file_name, "wb")
      for i in xrange(num_blocks):
        fh.write(block)
      fh.close()


# --- mini test ---
if __name__ == '__main__':
  import sys
  for a in sys.argv[1:]:
    # read image
    im = ImageFile(a)
    im.open()
    d = im.read_blk(0)
    im.write_blk(0,d)
    im.close()
    # read fobj
    fobj = file(a,"r+b")
    im = ImageFile(a,fobj=fobj)
    im.open()
    d = im.read_blk(0)
    im.write_blk(0,d)
    im.close()



import fsui
from launcher.res import gettext
from launcher.settings.settings_page import SettingsPage


class DirectoriesSettingsPage(SettingsPage):
    def __init__(self, parent):
        super().__init__(parent)
        icon = fsui.Icon("settings", "pkg:workspace")
        self.add_header(icon, gettext("Directories"))



import workspace
import workspace.ui


class BrowseButton(workspace.ui.ImageButton):
    def __init__(self, parent, url=""):
        super().__init__(
            parent,
            workspace.ui.Image(
                workspace.Stream(__name__, "data/16/folder.png")
            ),
        )
        self.url = url
        self.set_min_width(40)

    def on_activate(self):
        pass



from fsui.common.element import Element as Element
from fsui.common.element import LightElement as LightElement
from fsui.common.group import Group as Group
from fsui.common.layout import HorizontalLayout as HorizontalLayout
from fsui.common.layout import Layout as Layout
from fsui.common.layout import VerticalLayout as VerticalLayout
from fsui.common.spacer import Spacer as Spacer
from fsui.qt import Signal as Signal
from fsui.qt import get_screen_size as get_screen_size
from fsui.qt import show_error as show_error
from fsui.qt import call_after as call_after
from fsui.qt import call_later as call_later
from fsui.qt.adapter import Adapter as Adapter
from fsui.qt.Application import Application as Application
from fsui.qt.Button import Button as Button
from fsui.qt.CheckBox import CheckBox as CheckBox
from fsui.qt.CheckBox import HeadingCheckBox as HeadingCheckBox
from fsui.qt.Choice import Choice as Choice
from fsui.qt.Choice import ItemChoice as ItemChoice
from fsui.qt.Color import Color as Color
from fsui.qt.ComboBox import ComboBox as ComboBox
from fsui.qt.dialog import Dialog as Dialog
from fsui.qt.dialogbuttons import DialogButtons as DialogButtons
from fsui.qt.dialogwindow import DialogWindow as DialogWindow
from fsui.qt.DirDialog import DirDialog as DirDialog
from fsui.qt.DrawingContext import DrawingContext as DrawingContext
from fsui.qt.DrawingContext import Font as Font
from fsui.qt.FileDialog import FileDialog as FileDialog
from fsui.qt.FileDialog import pick_directory as pick_directory
from fsui.qt.FileDialog import pick_file as pick_file
from fsui.qt.FileDialog import pick_files as pick_files
from fsui.qt.HeadingLabel import HeadingLabel as HeadingLabel
from fsui.qt.Icon import Icon as Icon
from fsui.qt.Image import Image as Image
from fsui.qt.ImageButton import ImageButton as ImageButton
from fsui.qt.ImageView import ImageView as ImageView
from fsui.qt.label import Label as Label
from fsui.qt.label import MultiLineLabel as MultiLineLabel
from fsui.qt.label import PlainLabel as PlainLabel
from fsui.qt.label import URLLabel as URLLabel
from fsui.qt.legacydialog import LegacyDialog as LegacyDialog
from fsui.qt.ListView import ListView as ListView
from fsui.qt.Menu import Menu as Menu
from fsui.qt.Menu import PopupMenu as PopupMenu
from fsui.qt.Panel import Panel as Panel
from fsui.qt.scrollarea import VerticalScrollArea as VerticalScrollArea
from fsui.qt.SpinCtrl import SpinCtrl as SpinCtrl
from fsui.qt.System import System as System
from fsui.qt.TextArea import TextArea as TextArea
from fsui.qt.TextField import PasswordField as PasswordField
from fsui.qt.TextField import TextField as TextField
from fsui.qt.timer import IntervalTimer as IntervalTimer
from fsui.qt.VerticalItemView import VerticalItemView as VerticalItemView
from fsui.qt.window import Window as Window

default_window_parent = []
default_window_center = []
toolkit = "qt"
use_qt = True
theme = ""
theme_variant = ""


# noinspection PyProtectedMember
def open_window_instance(cls, parent=None):
    if not hasattr(cls, "_window_instance"):
        cls._window_instance = None
    if cls._window_instance is not None:
        cls._window_instance.raise_and_activate()
        return cls._window_instance
    cls._window_instance = cls(parent)

    def reset_instance():
        print("SettingsDialog.reset_instance")
        # cls._window_instance.deleteLater()
        cls._window_instance = None

    cls._window_instance.closed.connect(reset_instance)
    cls._window_instance.show()
    return cls._window_instance

    # def monitor_instance_2(count):
    #     if count < 100:
    #         call_after(monitor_instance_2, count + 1)
    #         return
    #     print("DESTROYED SIGNAL RECEIVED 2")
    #     instance = weak_instance()
    #     if instance is not None:
    #         print("WARNING: SettingsDialog is still alive")
    #         import gc
    #         print(gc.get_referrers(instance))
    #         print("real window:")
    #         print(gc.get_referrers(instance.real_window()))
    #     else:
    #         print("Instance is now", instance)
    #
    # def monitor_instance():
    #     print("DESTROYED SIGNAL RECEIVED")
    #     call_after(monitor_instance_2, 1)
    #
    # weak_instance = weakref.ref(cls._window_instance)
    # # cls._window_instance.destroyed.connect(monitor_instance)
    # cls._window_instance.closed.connect(monitor_instance)

    # if getattr(cls, "_window_instance", None) and cls._window_instance():
    #     cls._window_instance().raise_and_activate()
    # else:
    #     window = cls(parent)
    #     window.show()
    #     cls._window_instance = weakref.ref(window)


# noinspection PyProtectedMember
def current_window_instance(cls):
    if not hasattr(cls, "_window_instance"):
        return None
    return cls._window_instance



from fsgs.drivers.mess.messmsxdriver import MessMsxDriver
from fsgs.platform import PlatformHandler
from fsgs.platforms.loader import SimpleLoader


class MsxPlatformHandler(PlatformHandler):
    PLATFORM_NAME = "MSX"

    def __init__(self):
        PlatformHandler.__init__(self)

    def get_loader(self, fsgs):
        return MsxLoader(fsgs)

    def get_runner(self, fsgs):
        return MessMsxDriver(fsgs)


class MsxLoader(SimpleLoader):
    pass



import re
import subprocess
import traceback

from fsgs import Option
from fsgs.amiga.fsuaedevicehelper import FSUAEDeviceHelper
from launcher.devicemanager import DeviceManager

from .device import Device


class EnumerateHelper(object):
    def __init__(self):
        self.devices = []
        self.joystick_devices = []
        self.keyboard_devices = []
        self.joystick_like_devices = []
        self.initialized = False

    def update(self):
        # if cls.initialized:
        #     return
        # init can be called more than once (by setting initialized to
        # false, used by refresh function, so we need to clear the lists...
        self.devices = []
        self.joystick_devices = []
        self.keyboard_devices = []
        self.joystick_like_devices = []
        self.init_fsuae()

    def init(self):
        if self.initialized:
            return
        self.init_fsuae()
        self.initialized = True

    def init_fsuae(self):
        print("[INPUT] EnumerateHelper: Finding connected joysticks")
        try:
            p = FSUAEDeviceHelper.start_with_args(
                ["--list"], stdout=subprocess.PIPE
            )
            joysticks = p.stdout.read()
            p.wait()
        except Exception:
            print("[INPUT] Exception while listing joysticks and devices")
            traceback.print_exc()
            return
        print("[INPUT]", repr(joysticks))
        # If the character conversion fails, replace will ensure that
        # as much as possible succeeds. The joystick in question will
        # not be pre-selectable in the launcher, but the other ones will
        # work at least.
        joysticks = joysticks.decode("UTF-8", "replace")
        joysticks = [x.strip() for x in joysticks.split("\n") if x.strip()]

        # joysticks.append("J: controller xbox 360 for windows")
        # joysticks.append("Buttons: 10 Hats: 1 Axes: 5 Balls: 0")

        device_name_count = {}
        last_joystick = None

        for line in joysticks:
            if line.startswith("#"):
                continue
            elif line.startswith("Buttons:"):
                parts = line.split(" ")
                last_joystick.buttons = int(parts[1])
                last_joystick.hats = int(parts[3])
                last_joystick.axes = int(parts[5])
                last_joystick.balls = int(parts[7])
                continue
            elif line.startswith("SDLName:"):
                value = line.split(" ", 1)[1]
                # Strip quotes
                last_joystick.sdl_name = value[1:-1]
                continue

            device = Device()
            device_type, name = line.split(" ", 1)

            name_count = device_name_count.get((device_type, name), 0) + 1
            device_name_count[(device_type, name)] = name_count
            if name_count > 1:
                name = name + " #" + str(name_count)
            device.id = name
            name = re.sub("[ ]+", " ", name)
            device.name = name
            if device_type == "J:":
                device.type = "joystick"
                last_joystick = device
                self.joystick_devices.append(device)
            elif device_type == "M:":
                device.type = "mouse"
            elif device_type == "K:":
                # works as an emulated joystick...
                # device_type = "joystick"
                device.type = "keyboard"
                self.keyboard_devices.append(device)
            # self.device_types.append(device_type)
            self.devices.append(device)
        for i, device in enumerate(self.joystick_devices):
            device.index = i
        for i, device in enumerate(self.keyboard_devices):
            device.index = i

        self.joystick_like_devices.extend(self.joystick_devices)
        self.joystick_like_devices.extend(self.keyboard_devices)

    def default_port_selection(self, ports, options):
        print("[INPUT] Default port selection (EnumerateHelper)")
        self.init()
        # for device in self.devices:
        #     print(" #", device.id)
        #     device.configure("megadrive")

        if len(ports) > 0:
            if ports[0].type_option:
                print("[INPUT] New-style port device selection:")
                # Supports new-style port selection
                port_devices = DeviceManager.get_non_amiga_devices_for_ports(
                    options
                )
                from fsgs.platform import Platform

                if ports[0].type_option == Option.C64_PORT_2_TYPE:
                    print("[INPUT] Hack for inverted C64 port order")
                    temp = port_devices[1]
                    port_devices[1] = port_devices[2]
                    port_devices[2] = temp
                for i, port in enumerate(ports):
                    for device in self.devices:
                        if port_devices[i + 1].id == device.id:
                            port.device = device
                            print("[INPUT]", port.name, "<-", device.id)
                            break
                    else:
                        print("[INPUT]", port.name, "<- [None]")
                return

        joystick_like_devices = self.joystick_like_devices[:]
        print("[INPUT] Old-style port device selection:")
        for port in ports:
            for i, device in enumerate(joystick_like_devices):
                # device.configure()
                if True:
                    joystick_like_devices.pop(i)
                    port.device = device
                    break
            print("[INPUT] Old Selection:", port.name, port.device)



from __future__ import absolute_import
from __future__ import print_function

import time
from .Block import Block

class CommentBlock(Block):
  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num, is_type=Block.T_COMMENT)
    self.comment = ""

  def create(self, header_key, comment=""):
    Block.create(self)
    self.own_key = self.blk_num
    self.header_key = header_key
    self.comment = comment

  def set(self, data):
    self._set_data(data)
    self._read()

  def read(self):
    self._read_data()
    self._read()

  def _read(self):
    Block.read(self)
    if not self.valid:
      return False

    # Comment fields
    self.own_key = self._get_long(1)
    self.header_key = self._get_long(2)
    self.checksum = self._get_long(5)
    self.comment = self._get_bstr(6, 79)
    self.valid = (self.own_key == self.blk_num)
    return self.valid

  def write(self):
    Block._create_data(self)
    self._put_long(1, self.own_key)
    self._put_long(2, self.header_key)
    self._put_bstr(6, 79, self.comment)
    Block.write(self)

  def dump(self):
    Block.dump(self,"Comment")
    print(" own_key:    %d" % (self.own_key))
    print(" header_key: %d" % (self.header_key))
    print(" comment:    '%s'" % self.comment)



from __future__ import absolute_import
from __future__ import print_function

from .BlockScan import BlockScan
from amitools.fs.FSString import FSString
from amitools.fs.FileName import FileName
from amitools.fs.validate.Log import Log
import amitools.fs.DosType as DosType

class FileData:
  def __init__(self, bi):
    self.bi = bi

class FileInfo:
  def __init__(self, bi):
    self.bi = bi

class FileScan:
  def __init__(self, block_scan, log, dos_type):
    self.block_scan = block_scan
    self.log = log
    self.dos_type = dos_type
    self.ffs = DosType.is_ffs(self.dos_type)
    self.infos = []
  
  def scan_all_files(self, all_file_hdr_block_infos, progress=None):
    """scan through all files"""
    if progress != None:
      progress.begin("file")
      for bi in all_file_hdr_block_infos:
        self.scan_file(bi)
        progress.add()
      progress.end()
    else:
      for bi in all_file_hdr_block_infos:
        self.scan_file(bi)
        
  def scan_file(self, bi):
    """scan a file header block info and create a FileInfo instance"""
    fi = FileInfo(bi)
    self.infos.append(fi)
    
    info = "'%s' (@%d)" % (bi.name, bi.blk_num)
    
    # scan for data blocks
    linked_data_blocks = bi.data_blocks
    blk_num = bi.blk_num
    # run through file list blocks linked by extension
    sbi = bi
    aborted = False
    num = 0
    while sbi.extension != 0 and sbi.extension < self.block_scan.blkdev.num_blocks:
      # check usage of block
      if self.block_scan.is_block_available(sbi.extension):
        self.log.msg(Log.ERROR, "File ext block #%d of %s already used" % (num, info), sbi.extension)
        aborted = True
        break
      # get block
      ebi = self.block_scan.get_block(sbi.extension)
      if ebi == None:
        aborted = True
        break
      # check block type
      if ebi.blk_type != BlockScan.BT_FILE_LIST:
        self.log.msg(Log.ERROR, "File ext block #%d of %s is no ext block" % (num, info), ebi.blk_num)
        aborted = True
        break
      # check for parent link
      if ebi.parent_blk != blk_num:
        self.log.msg(Log.ERROR, "File ext block #%d of %s has invalid parent: got %d != expect %d" % (num, info, ebi.parent_blk, blk_num), ebi.blk_num)
      # warn if data blocks is not full
      ndb = len(ebi.data_blocks)
      if ebi.extension != 0 and ndb != self.block_scan.blkdev.block_longs - 56:
        self.log.msg(Log.WARN, "File ext block #%d of %s has incomplete data refs: got %d" % (num, info, ndb), ebi.blk_num)
      # add data blocks
      linked_data_blocks += ebi.data_blocks
      sbi = ebi
      num += 1
    
    # transform the data block numbers to file data
    file_datas = []
    seq_num = 1
    for data_blk in linked_data_blocks:
      # get block
      block_used = self.block_scan.is_block_available(data_blk)
      dbi = self.block_scan.get_block(data_blk)
      fd = FileData(dbi)
      file_datas.append(fd)
      # check usage of block
      # is block available
      if dbi == None:
        self.log.msg(Log.ERROR, "File data block #%d of %s not found" % (seq_num,info), data_blk)
      if block_used:
        self.log.msg(Log.ERROR, "File data block #%d of %s already used" % (seq_num,info), data_blk)
        fd.bi = None
      # in ofs check data blocks
      if not self.ffs:
        # check block type
        if dbi.blk_type != BlockScan.BT_FILE_DATA:
          self.log.msg(Log.ERROR, "File data block #%d of %s is no data block" % (seq_num,info), data_blk)
          fd.bi = None
        else:
          # check header ref: must point to file header
          if dbi.hdr_key != blk_num:
            self.log.msg(Log.ERROR, "File data block #%d of %s does not ref header: got %d != expect %d" % (seq_num,info,dbi.hdr_key, blk_num), data_blk)          
          # check sequence number
          if dbi.seq_num != seq_num:
            self.log.msg(Log.ERROR, "File data block #%d of %s seq num mismatch: got %d" % (seq_num,info,dbi.seq_num), data_blk)
      seq_num += 1
    
    # check size of file in bytes
    block_data_bytes = self.block_scan.blkdev.block_bytes
    if not self.ffs:
      block_data_bytes -= 24
    file_est_blocks = (bi.byte_size + block_data_bytes - 1) / block_data_bytes
    num_data_blocks = len(linked_data_blocks)
    if file_est_blocks != num_data_blocks:
      self.log.msg(Log.ERROR, "File %s with %d bytes has wrong number of data blocks: got %d != expect %d" % (info, bi.byte_size, num_data_blocks, file_est_blocks), bi.blk_num)
    
    return fi
  
  def dump(self):
    pass
    


from __future__ import absolute_import
from __future__ import print_function

from .BlockDevice import BlockDevice
import os.path
import os

from .ImageFile import ImageFile

class RawBlockDevice(BlockDevice):
  def __init__(self, raw_file, read_only=False, block_bytes=512, fobj=None):
    self.img_file = ImageFile(raw_file, read_only, block_bytes, fobj)

  def create(self, num_blocks):
    self.img_file.create(num_blocks)
    self.open()
    self.num_blocks = num_blocks

  def open(self):
    self.img_file.open()
    # calc block longs
    self.block_bytes = self.img_file.block_bytes
    self.block_longs = self.block_bytes / 4
    self.num_blocks = self.img_file.num_blocks

  def flush(self):
    self.img_file.flush()

  def close(self):
    self.img_file.close()

  def read_block(self, blk_num):
    return self.img_file.read_blk(blk_num)

  def write_block(self, blk_num, data):
    self.img_file.write_blk(blk_num, data)



import fsui
from fsbc.application import app
from launcher.i18n import gettext
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior


class FullscreenModeButton(fsui.ImageButton):
    def __init__(self, parent):
        self.window_icon = fsui.Image(
            "launcher:res/16x16/fullscreen_window.png"
        )
        self.fullscreen_icon = fsui.Image(
            "launcher:res/16x16/fullscreen_fullscreen.png"
        )
        self.desktop_icon = fsui.Image(
            "launcher:res/16x16/fullscreen_desktop.png"
        )
        super().__init__(parent, self.desktop_icon)
        self.set_tooltip(
            gettext("Change fullscreen mode (desktop, fullscreen, window)")
        )
        self.set_min_width(40)
        self.fullscreen_mode = "desktop"
        SettingsBehavior(self, ["fullscreen", "fullscreen_mode"])

    def on_fullscreen_setting(self, value):
        # self.set_enabled(value == "1")
        pass

    def on_fullscreen_mode_setting(self, value):
        if value == "fullscreen":
            self.fullscreen_mode = "fullscreen"
            self.set_image(self.fullscreen_icon)
        elif value == "window":
            self.fullscreen_mode = "window"
            self.set_image(self.window_icon)
        else:
            self.fullscreen_mode = "desktop"
            self.set_image(self.desktop_icon)

    def on_activate(self):
        if self.fullscreen_mode == "fullscreen":
            app.settings["fullscreen_mode"] = "window"
        elif self.fullscreen_mode == "window":
            app.settings["fullscreen_mode"] = ""
        else:
            app.settings["fullscreen_mode"] = "fullscreen"



from fsgs.ogd.client import OGDClient
import fsui
from fsbc.application import app

# from workspace.shell import SimpleApplication
from launcher.res import gettext
from launcher.ui.widgets import CloseButton
from workspace.ui.theme import WorkspaceTheme


class LogoutWindow(fsui.Window):
    @classmethod
    def open(cls, parent=None):
        return fsui.open_window_instance(cls, parent)

    def __init__(self, parent=None):
        title = gettext("Log Out from Your OpenRetro Account")
        super().__init__(parent, title, minimizable=False, maximizable=False)
        self.theme = WorkspaceTheme.instance()
        self.layout = fsui.VerticalLayout()
        self.set_icon(fsui.Icon("password", "pkg:workspace"))

        self.layout.set_padding(20, 20, 20, 20)

        heading_layout = fsui.HorizontalLayout()
        self.layout.add(heading_layout)
        heading_layout.add(
            fsui.ImageView(self, fsui.Image("workspace:res/48/password.png"))
        )
        heading_layout.add_spacer(20)
        heading_layout_2 = fsui.VerticalLayout()
        heading_layout.add(
            heading_layout_2, expand=True, fill=False, valign=0.5
        )
        heading_layout_2.add(
            fsui.HeadingLabel(
                self, gettext("Log Out from Your OpenRetro Account")
            )
        )
        heading_layout_2.add_spacer(2)
        heading_layout_2.add(
            fsui.Label(
                self,
                gettext("While logged out you will not get database updates"),
            )
        )

        self.layout.add_spacer(20)
        hori_layout = fsui.HorizontalLayout()
        self.layout.add(hori_layout, fill=True)
        self.created_label = fsui.Label(self, "")
        hori_layout.add(self.created_label, expand=True)
        hori_layout.add_spacer(20)

        self.logout_button = fsui.Button(self, gettext("Log Out"))
        # self.logout_button.disable()
        self.logout_button.activated.connect(self.__logout_activated)
        hori_layout.add(self.logout_button)

        if self.window.theme.has_close_buttons:
            self.close_button = CloseButton(self)
            hori_layout.add(self.close_button, fill=True, margin_left=10)

    def __del__(self):
        print("LogoutWindow.__del__")

    def __logout_activated(self):
        auth_token = app.settings["database_auth"]
        if auth_token:
            task = OGDClient().logout_task(auth_token)
            # task.progressed.connect(self.progress)
            task.succeeded.connect(self.close)
            # task.failed.connect(fsui.error_function(gettext("Login Failed")))
            task.failed.connect(self.on_failure)
            task.start()
        else:
            # this is not a normal case, no auth token stored, but clear
            # all auth-related settings just in case
            app.settings["database_auth"] = ""
            app.settings["database_username"] = ""
            # app.settings["database_email"] = ""
            app.settings["database_password"] = ""
            self.on_close()

    def on_failure(self, message):
        fsui.show_error(message, parent=self.window)

    # def on_progress(self):
    #    print("on_progress")

    # def on_success(self):
    #     app.settings["database_auth"] = ""
    #     app.settings["database_username"] = ""
    #     app.settings["database_email"] = ""
    #     app.settings["database_password"] = ""
    #     self.close()


# application = SimpleApplication(LogoutWindow)



from amitools.fs.block.rdb.PartitionBlock import *
from amitools.fs.blkdev.PartBlockDevice import PartBlockDevice
import amitools.util.ByteSize as ByteSize
import amitools.fs.DosType as DosType

class Partition:
  def __init__(self, blkdev, blk_num, num, cyl_blks, rdisk):
    self.blkdev = blkdev
    self.blk_num = blk_num
    self.num = num
    self.cyl_blks = cyl_blks
    self.rdisk = rdisk
    self.part_blk = None

  def get_next_partition_blk(self):
    if self.part_blk != None:
      return self.part_blk.next
    else:
      return 0xffffffff

  def get_blk_num(self):
    """return the block number of the partition block"""
    return self.blk_num

  def read(self):
    # read fs header
    self.part_blk = PartitionBlock(self.blkdev, self.blk_num)
    if not self.part_blk.read():
      self.valid = False
      return False
    self.valid = True
    return True

  def create_blkdev(self, auto_close_rdb_blkdev=False):
    """create a block device for accessing this partition"""
    return PartBlockDevice(self.blkdev, self.part_blk, auto_close_rdb_blkdev)

  def write(self):
    self.part_blk.write()

  # ----- Query -----

  def dump(self):
    self.part_blk.dump()

  def get_num_blocks(self):
    """return total number of blocks in this partition"""
    p = self.part_blk
    cyls = p.dos_env.high_cyl - p.dos_env.low_cyl + 1
    return cyls * self.cyl_blks

  def get_num_bytes(self, block_size=512):
    return self.get_num_blocks() * block_size

  def get_drive_name(self):
    return self.part_blk.drv_name

  def get_flags(self):
    return self.part_blk.flags

  def get_index(self):
    return self.num

  def get_cyl_range(self):
    de = self.part_blk.dos_env
    if de == None:
      return None
    else:
      return (de.low_cyl, de.high_cyl)

  def get_info(self, total_blks=0):
    """return a string line with typical info about this partition"""
    p = self.part_blk
    de = p.dos_env
    name = "'%s'" % p.drv_name
    part_blks = self.get_num_blocks()
    part_bytes = self.get_num_bytes()
    extra = ""
    if total_blks != 0:
      ratio = 100.0 * part_blks / total_blks
      extra += "%6.2f%%  " % ratio
    # add dos type
    dos_type = de.dos_type
    extra += DosType.num_to_tag_str(dos_type)
    extra += "/0x%04x" % dos_type
    # max transfer
    extra += " max_transfer=0x%x" % de.max_transfer
    extra += " mask=0x%x" % de.mask
    extra += " num_buffer=%d" % de.num_buffer
    # add flags
    flags = p.flags
    if flags & PartitionBlock.FLAG_BOOTABLE == PartitionBlock.FLAG_BOOTABLE:
      extra += " bootable pri=%d" % de.boot_pri
    if flags & PartitionBlock.FLAG_NO_AUTOMOUNT == PartitionBlock.FLAG_NO_AUTOMOUNT:
      extra += " no_automount"
    return "Partition: #%d %-06s %8d %8d  %10d  %s  %s" \
      % (self.num, name, de.low_cyl, de.high_cyl, part_blks, ByteSize.to_byte_size_str(part_bytes), extra)





import sys

import fsui
from fsbc.application import app
from fsgs.drivers.gamedriver import GameDriver
from launcher.i18n import gettext
from launcher.launcher_config import LauncherConfig
from launcher.launcher_settings import LauncherSettings
from launcher.option import Option
from launcher.settings.override_warning import OverrideWarning
from launcher.ui.behaviors.settingsbehavior import SettingsBehavior


class LaunchGroup(fsui.Group):
    def __init__(self, parent, gsc):
        super().__init__(parent)
        self.layout = fsui.VerticalLayout()
        self.hori_layout = fsui.HorizontalLayout()
        self.layout.add(self.hori_layout, fill=True, expand=True)
        self.progress_label = fsui.Label(self, "")
        self.progress_label.set_visible(False)
        self.hori_layout.add(self.progress_label, fill=True, expand=True)

        # self.progress_2_label = fsui.Label(self, "")
        # self.progress_2_label.set_visible(False)
        # self.hori_layout.add(self.progress_2_label, fill=True, expand=True)

        # self.fullscreen_mode_button = FullscreenModeButton(self)
        # self.hori_layout.add(
        #     self.fullscreen_mode_button, fill=True, margin_right=10)
        # self.monitor_button = MonitorButton(self)
        # self.hori_layout.add(
        #     self.monitor_button, fill=True, margin_right=10)

        self.screen_info_label = ScreenInfoLabel(self)
        self.hori_layout.add(
            self.screen_info_label, fill=True, expand=True, margin_right=10
        )
        # self.video_sync_checkbox = VideoSyncCheckBox(self)
        # self.hori_layout.add(self.video_sync_checkbox, margin_right=10)
        self.override_warning = OverrideWarning(self, Option.FULLSCREEN)
        self.hori_layout.add(self.override_warning, margin_right=10)
        self.fullscreen_button = FullscreenToggleButton(self)
        self.hori_layout.add(self.fullscreen_button, fill=True)
        start_button = StartButton(self, gsc)
        self.hori_layout.add(start_button, fill=True, margin_left=10)
        # ConfigBehavior(self, [Option.FULLSCREEN])
        gsc.config.add_behavior(
            self, [Option.FULLSCREEN, "__running", "__progress"]
        )

    def on___running_config(self, value):
        widgets = [
            # self.fullscreen_mode_button,
            # self.monitor_button,
            self.screen_info_label,
            # self.video_sync_checkbox,
            self.fullscreen_button,
        ]
        for widget in widgets:
            widget.set_visible(value != "1")
        self.progress_label.set_visible(value == "1")
        # self.progress_2_label.set_visible(value == "1")
        self.hori_layout.update()

    def on___progress_config(self, value):
        # first, rest = value.split(":", 1)
        # self.progress_label.set_text(first)
        # self.progress_2_label.set_text(rest.strip())
        self.progress_label.set_text(value)

    def on_fullscreen_config(self, _):
        self.layout.update()


class StartButton(fsui.Button):
    def __init__(self, parent, gsc):
        super().__init__(parent, gettext("Start"))
        self.gsc = gsc
        #        gsc.signal.add_behavior(self, ["start_ready"])
        gsc.config.add_behavior(self, ["__running"])

    def on_activated(self):
        from launcher.launcherapp import LauncherApp

        LauncherApp.start_game()

    def on___running_config(self, value):
        self.set_enabled(value != "1")

        # def on_start_ready_signal(self):
        #     pass


class ScreenInfoLabel(fsui.Label):
    def __init__(self, parent):
        super().__init__(parent, "")
        try:
            from fsui.qt import init_qt

            qapplication = init_qt()
        except AttributeError:
            pass
        else:
            for screen in qapplication.screens():
                screen.geometryChanged.connect(self.on_screen_change)
                screen.refreshRateChanged.connect(self.on_screen_change)
            qapplication.screenAdded.connect(self.on_screen_added)
        SettingsBehavior(
            self, ["fullscreen", "monitor", "assume_refresh_rate"]
        )

    def on_fullscreen_setting(self, _):
        self.update_info()

    def on_screen_change(self, _):
        print("-- screen changed --")
        self.update_info()

    def on_screen_added(self, screen):
        print("-- screen added --")
        screen.geometryChanged.connect(self.on_screen_change)
        screen.refreshRateChanged.connect(self.on_screen_change)
        self.update_info()

    def on_assume_refresh_rate_setting(self, _):
        self.update_info()

    def on_monitor_setting(self, value):
        self.update_info(value=value)

    def update_info(self, value=None):
        if LauncherSettings.get("fullscreen") != "1":
            self.set_text("")
            return
        if value is None:
            value = LauncherSettings.get("monitor")
        x, y, w, h = GameDriver.screen_rect_for_monitor(value)
        # index = GameRunner.screen_index_for_monitor(value)
        refresh_rate = GameDriver.screen_refresh_rate_for_monitor(value)
        assume_refresh_rate = LauncherSettings.get("assume_refresh_rate")
        if assume_refresh_rate:
            refresh_rate_override = " (OVERRIDING {})".format(refresh_rate)
            try:
                refresh_rate = float(assume_refresh_rate)
            except ValueError:
                refresh_rate = 0.0
        else:
            refresh_rate_override = ""
        if x or y:
            pos_str = " ({}, {})".format(x, y)
        else:
            pos_str = ""
        self.set_text(
            "{}x{}{} @ {}Hz{}".format(
                w, h, pos_str, refresh_rate, refresh_rate_override
            )
        )


class FullscreenToggleButton(fsui.ImageButton):
    def __init__(self, parent):
        self.windowed_icon = fsui.Image("launcher:res/windowed_16.png")
        self.fullscreen_icon = fsui.Image("launcher:res/fullscreen_16.png")
        fsui.ImageButton.__init__(self, parent, self.windowed_icon)
        self.set_tooltip(
            gettext("Toggle Between Windowed and Full-Screen Mode")
        )
        self.set_min_width(40)
        self.fullscreen_mode = False
        self.on_setting("fullscreen", app.settings["fullscreen"])
        LauncherSettings.add_listener(self)

    def on_destroy(self):
        LauncherSettings.remove_listener(self)

    def on_setting(self, key, value):
        if key == "fullscreen":
            if value == "1":
                self.fullscreen_mode = True
                self.set_image(self.fullscreen_icon)
            else:
                self.fullscreen_mode = False
                self.set_image(self.windowed_icon)

    def on_activate(self):
        if self.fullscreen_mode:
            app.settings["fullscreen"] = "0"
        else:
            app.settings["fullscreen"] = "1"


class LaunchDialog(fsui.Window):
    def __init__(self, parent, title, task):
        print("LaunchDialog parent =", parent)
        self.has_parent = parent is not None
        self.no_gui = "--no-gui" in sys.argv
        super().__init__(parent, title, maximizable=False)
        self.layout = fsui.VerticalLayout()

        self.layout.add_spacer(400, 20)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True)

        hor_layout.padding_right = 20
        hor_layout.add_spacer(20)

        image = fsui.Image("launcher:res/fs_uae_group.png")
        self.image_view = fsui.ImageView(self, image)
        hor_layout.add(self.image_view, valign=0.0)
        hor_layout.add_spacer(20)

        ver_layout = fsui.VerticalLayout()
        hor_layout.add(ver_layout, fill=True, expand=True)
        self.title_label = fsui.HeadingLabel(self, title)
        ver_layout.add(self.title_label, fill=True)

        ver_layout.add_spacer(6)
        self.sub_title_label = fsui.Label(self, gettext("Preparing..."))
        ver_layout.add(self.sub_title_label, fill=True)

        self.layout.add_spacer(20)

        hor_layout = fsui.HorizontalLayout()
        self.layout.add(hor_layout, fill=True)

        hor_layout.add_spacer(20, expand=True)
        self.cancel_button = fsui.Button(self, gettext("Cancel"))
        self.cancel_button.activated.connect(self.on_cancel_button)
        hor_layout.add(self.cancel_button)
        hor_layout.add_spacer(20)

        self.layout.add_spacer(20)
        self.set_size(self.layout.get_min_size())
        self.center_on_parent()

        self.was_closed = False
        self.task = task
        self.task.progressed.connect(self.on_progress)
        self.task.finished.connect(self.on_complete)
        self.task.failed.connect(self.on_error)

        self.closed.connect(self.__closed)

    def complete(self):
        print("TaskRunner.complete")
        # Setting task to None so destructors are run now
        self.task = None
        self.was_closed = True
        self.close()
        self.deleteLater()

    def __closed(self):
        LauncherConfig.set("__running", "")
        self.cancel()
        return False

    def on_progress(self, progress):
        def hide_function():
            self.visible = False

        def function():
            if progress == "__run__":
                self.cancel_button.disable()
                # Hide dialog after 1.5 seconds. The reason for delaying it
                # is to avoid "confusing" flickering if/when the dialog is
                # only shown for a split second.
                if self.is_shown():
                    fsui.call_later(1500, hide_function)
                LauncherConfig.set("__progress", gettext("Running: Emulator"))
            else:
                if self.no_gui:
                    print("[PROGRESS]", progress)
                else:
                    self.sub_title_label.set_text(progress)
                LauncherConfig.set(
                    "__progress", "Preparing: {}".format(progress)
                )

        fsui.call_after(function)

    def show(self, *args, **kwargs):
        if self.has_parent or self.no_gui:
            # Hack to prevent it from being shown
            pass
        else:
            self.center_on_screen()
            super().show()

    def on_complete(self):
        def function():
            self.complete()

        fsui.call_after(function)

    def on_error(self, message):
        fsui.show_error(message, title="Guru Meditation")
        self.close()

    def on_cancel_button(self):
        self.cancel()

    def cancel(self):
        print("LaunchDialog.cancel")
        if self.task is not None:
            self.task.stop()
        self.cancel_button.disable()



from fsbc.desktop import open_url_in_browser
from fsgs.network import openretro_url_prefix
from fsui import Image, Panel

from ...launcher_config import LauncherConfig
from ..skin import Skin


class WebButton(Panel):
    def __init__(self, parent, icon=None):
        if icon is not None:
            self.icon = icon
        else:
            self.icon = Image("launcher:res/16x16/world.png")
        Panel.__init__(self, parent, paintable=True)
        # self.set_tooltip(tooltip)
        LauncherConfig.add_listener(self)
        self.on_config("variant_uuid", "")

    def on_destroy(self):
        LauncherConfig.remove_listener(self)
        pass

    def get_min_width(self):
        return 32

    def get_min_height(self):
        return 24

    def get_url(self):
        variant_uuid = LauncherConfig.get("variant_uuid", "")
        if not variant_uuid:
            return
        return "{0}/game/{1}".format(openretro_url_prefix(), variant_uuid)

    def on_left_down(self):
        url = self.get_url()
        if url:
            open_url_in_browser(url)

    def on_config(self, key, value):
        if key == "variant_uuid":
            if value:
                if not self.is_enabled():
                    self.enable()
                    # self.refresh()
                    self.show()
                    self.set_hand_cursor()
            else:
                if self.is_enabled():
                    self.disable()
                    self.hide()
                    # self.refresh()
                    # self.set_default_cursor()

    def on_paint(self):
        dc = self.create_dc()
        dc.clear(Skin.get_background_color())
        if self.is_enabled():
            dc.draw_image(self.icon, 8, 4)
        else:
            pass



import platform
from os import path

import fsboot
from fsbc.system import System

X86_MACHINES = ["x86", "i386", "i486", "i586", "i686"]
X86_64_MACHINES = ["x86_64", "x86-64", "amd64"]
X86_ANY_MACHINES = X86_MACHINES + X86_64_MACHINES

# Mapping between executable name and plugin
known_executables = {
    "dosbox": "DOSBox",
    "dosbox-staging": "DOSBox-Staging",
    "fs-fuse": "FS-Fuse",
    "fs-uae": "FS-UAE",
    "fs-uae-device-helper": "FS-UAE",
    "fuse": "Fuse",
    "hatari": "Hatari",
    "mame": "MAME",
    "mednafen": "Mednafen",
    "x64sc": "Vice",
}


class PluginExecutableFinder:
    def find_executable(self, name: str):
        print("Find executable:", name)
        exe_file = find_executable(name)
        if exe_file:
            exe_file = path.normpath(exe_file)
            print("->", exe_file)
            return exe_file
        print("-> None")
        return None


def find_executable(name: str):
    if fsboot.development():
        exe_file = find_executable_in_development_project_dir(name)
        if exe_file:
            return exe_file
    exe_file = find_executable_in_plugins_dir(name)
    if exe_file:
        return exe_file
    if System.macos and fsboot.is_frozen():
        exe_file = find_executable_in_side_by_side_app_bundle(name)
        if exe_file:
            return exe_file
    exe_file = find_executable_side_by_side(name)
    if exe_file:
        return exe_file
    exe_file = find_executable_in_side_by_side_plugin(name)
    if exe_file:
        return exe_file
    return None


def find_executable_in_development_project_dir(name: str):
    print("- Find executable in development project dir")
    plugin_name = known_executables.get(name)
    if plugin_name is None:
        return None
    project_dir_name = plugin_name.lower()
    exe_dir = fsboot.executable_dir()
    if exe_dir.endswith("-private"):
        exe_file = path.join(
            exe_dir,
            "..",
            f"{project_dir_name}-private",
            get_exe_name(name),
        )
        if check_executable(exe_file):
            return exe_file
        if project_dir_name == "fs-uae":
            exe_file = path.join(
                exe_dir,
                "..",
                "fs-uae-3-private",
                get_exe_name(name),
            )
            if check_executable(exe_file):
                return exe_file
    exe_file = path.join(exe_dir, "..", project_dir_name, get_exe_name(name))
    if check_executable(exe_file):
        return exe_file
    return None


def find_executable_in_plugins_dir(name: str):
    print("- Find executable in plugins dir")
    plugin_name = known_executables.get(name)
    if plugin_name is None:
        return None
    base_dir = fsboot.base_dir()
    plugins_dir = path.join(base_dir, "System")
    exe_file = find_executable_in_dir_containing_plugin(
        name, plugins_dir, plugin_name
    )
    if exe_file:
        return exe_file
    print("  Looking in legacy plugin directories")
    plugins_dir = path.join(base_dir, "Plugins")
    exe_file = find_executable_in_dir_containing_plugin(
        name, plugins_dir, plugin_name
    )
    if exe_file:
        return exe_file
    plugins_dir = path.join(base_dir, "Data", "Plugins")
    exe_file = find_executable_in_dir_containing_plugin(
        name, plugins_dir, plugin_name
    )
    if exe_file:
        return exe_file
    return None


def find_executable_in_side_by_side_app_bundle(name: str):
    print("- Find executable in side-by-side app bundle")
    plugin_name = known_executables.get(name)
    if plugin_name is None:
        return None
    exe_file = path.join(
        fsboot.app_dir(), f"{plugin_name}.app", "Contents", "MacOS", name
    )
    if check_executable(exe_file):
        return exe_file


def find_executable_in_side_by_side_plugin(name: str):
    print("- Find executable in side-by-side plugin")
    plugin_name = known_executables.get(name)
    if plugin_name is None:
        return None
    # plugin_dir = fsboot.plugin_dir()
    # if plugin_dir is None:
    #     print("  Not running in a plugin")
    #     return None
    # plugins_dir = path.join(plugin_dir, "..")
    plugins_dir = path.join(fsboot.app_dir(), "..", "..", "..")
    return find_executable_in_dir_containing_plugin(
        name, plugins_dir, plugin_name
    )


def find_executable_in_dir_containing_plugin(
    name: str, plugins_dir: str, plugin_name: str
):
    bin_dir = path.join(plugins_dir, plugin_name, os_name(), arch_name())
    return find_executable_in_dir_or_app(name, bin_dir, plugin_name)


def find_executable_in_dir_or_app(name: str, dir: str, plugin_name: str):
    if System.macos:
        exe_file = path.join(
            dir, f"{plugin_name}.app", "Contents", "MacOS", name
        )
        if check_executable(exe_file):
            return exe_file
    else:
        exe_file = path.join(dir, get_exe_name(name))
        if check_executable(exe_file):
            return exe_file
    return None


def find_executable_side_by_side(name: str):
    """Find executable side by side, for example in /usr/bin or similar."""
    print("- Find executable side-by-side")
    exe_file = path.join(fsboot.executable_dir(), get_exe_name(name))
    if check_executable(exe_file):
        return exe_file


def get_exe_name(name: str) -> str:
    if System.windows:
        return f"{name}.exe"
    return name


def os_name() -> str:
    if System.windows:
        return "Windows"
    elif System.macos:
        return "macOS"
    elif System.linux:
        return "Linux"
    else:
        return "Other"


def arch_name() -> str:
    if platform.machine().lower() in X86_ANY_MACHINES:
        if platform.architecture()[0] == "64bit":
            return "x86-64"
        else:
            return "x86"
    # FIXME: arm7/8 and ARM64
    return "Unknown"


def check_executable(exe_file: str) -> bool:
    if path.exists(exe_file):
        print(f"  Check executable {exe_file}: YES")
        return True
    else:
        print(f"  Check executable {exe_file}: NO")
        return False



# amitools utility functions

from __future__ import print_function


def _get_vis_char(d):
  v = ord(d)
  if v >= 32 and v < 127:
    return "%c" % d
  else:
    return "."

def get_hex_line(addr, line, indent=0, num=16):
  l = len(line)
  skip = num - l
  out = " " * indent
  out += "%08x: " % addr
  for d in line:
    out += "%02x " % ord(d)
  for d in xrange(skip):
    out += "   "
  out += " "
  for d in line:
    out += _get_vis_char(d)
  return out

def print_hex(data, indent=0, num=16, out=print, base_addr=0):
  l = len(data)
  o = 0
  addr = base_addr
  while o < l:
    line = data[o:o+num]
    out(get_hex_line(addr, line, indent, num))
    o += num
    addr += num

def get_hex_diff_line(addr, a_line, b_line, indent=0, num=16):
  na = len(a_line)
  nb = len(b_line)
  n = max(na, nb)
  skip = num - n
  out = " " * indent
  out += "%08x: " % addr
  ah = []
  ac = []
  bh = []
  bc = []
  for d in xrange(n):
    av = a_line[d]
    bv = b_line[d]
    if av != bv:
      ah.append("%02x" % ord(av))
      ac.append(_get_vis_char(av))
      bh.append("%02x" % ord(bv))
      bc.append(_get_vis_char(bv))
    else:
      ah.append("--")
      ac.append(" ")
      bh.append("--")
      bc.append(" ")
  for d in xrange(skip):
    ah.append("  ")
    ac.append(" ")
    bh.append("  ")
    bc.append(" ")
  out += " ".join(ah) + "  " + "".join(ac) + " | "
  out += " ".join(bh) + "  " + "".join(bc)
  return out

def print_hex_diff(a_data, b_data, indent=0, num=16, out=print,
                   show_same=False, base_addr=0):
  na = len(a_data)
  nb = len(b_data)
  n = max(na, nb)
  o = 0
  addr = base_addr
  while o < n:
    a_line = a_data[o:o+num]
    b_line = b_data[o:o+num]
    if show_same or a_line != b_line:
      out(get_hex_diff_line(addr, a_line, b_line, indent, num))
    o += num
    addr += num


# mini test
if __name__ == '__main__':
  print_hex("hello, world!", num=8, base_addr=0xf80000)
  print_hex_diff("hello, world!", "hello. WOrld!", num=8, base_addr=0xf80000)



#!/usr/bin/env python2.7
# rdbtool
# swiss army knife for rdb disk images or devices

from __future__ import absolute_import
from __future__ import print_function

import sys
import argparse
import os.path

from amitools.util.CommandQueue import CommandQueue
from amitools.fs.FSError import FSError
from amitools.fs.rdb.RDisk import RDisk
from amitools.fs.blkdev.RawBlockDevice import RawBlockDevice
from amitools.fs.blkdev.DiskGeometry import DiskGeometry
from amitools.fs.DosType import *
from amitools.fs.block.rdb.PartitionBlock import PartitionBlock, PartitionDosEnv
from amitools.fs.block.rdb.FSHeaderBlock import FSHeaderDeviceNode
import amitools.util.KeyValue as KeyValue
import amitools.util.ByteSize as ByteSize
import amitools.util.VerTag as VerTag

# ----- commands -----
class Command:
  def __init__(self, args, opts, edit=False):
    self.args = args
    self.opts = opts
    self.edit = edit
    self.exit_code = 0
    self.blkdev = None
    self.rdisk = None

  def run(self, blkdev, rdisk):
    # optional init blkdev function
    if hasattr(self, "init_blkdev"):
      if blkdev == None:
        self.blkdev = self.init_blkdev(self.args.image_file)
        if self.blkdev == None:
          return 5
        blkdev = self.blkdev

    # optional init rdisk function
    if hasattr(self, "init_rdisk"):
      # close old
      if rdisk != None:
        rdisk.close()
      # create new rdisk
      self.rdisk = self.init_rdisk(blkdev)
      if self.rdisk == None:
        return 6
      rdisk= self.rdisk

    # common handler
    if hasattr(self, 'handle_blkdev'):
      return self.handle_blkdev(blkdev)
    elif hasattr(self, 'handle_rdisk'):
      return self.handle_rdisk(rdisk)
    else:
      return 0

  def has_init_blkdev(self):
    return hasattr(self, 'init_blkdev')

  def need_rdisk(self):
    return hasattr(self, 'handle_rdisk') and not hasattr(self, 'init_rdisk')

class FSCommandQueue(CommandQueue):
  def __init__(self, args, cmd_list, sep, cmd_map):
    CommandQueue.__init__(self, cmd_list, sep, cmd_map)
    self.args = args
    self.blkdev = None
    self.rdisk = None

  def run(self):
    self.img = self.args.image_file
    try:
      # main command loop
      exit_code = CommandQueue.run(self)
    except FSError as e:
      cmd = "'%s'" % " ".join(self.cmd_line)
      print(cmd, "FSError:", str(e))
      exit_code = 3
    except IOError as e:
      cmd = "'%s'" % " ".join(self.cmd_line)
      print(cmd, "IOError:", str(e))
      exit_code = 4
    finally:
      # close rdisk
      if self.rdisk != None:
        self.rdisk.close()
        if self.args.verbose:
          print("closing rdisk:", self.img)
      # close blkdev
      if self.blkdev != None:
        self.blkdev.close()
        if self.args.verbose:
          print("closing image:", self.img)
    return exit_code

  def create_cmd(self, cclass, name, opts):
    return cclass(self.args, opts)

  def _open_rdisk(self):
    if self.rdisk == None:
      self.rdisk = RDisk(self.blkdev)
      if self.args.verbose:
        print("opening rdisk:", self.img)
      return self.rdisk.open()
    else:
      return True

  def run_first(self, cmd_line, cmd):
    self.cmd_line = cmd_line

    # check if first command is an init command
    if not cmd.has_init_blkdev():
      # auto add 'open' command
      pre_cmd = OpenCommand(self.args, [])
      if self.args.verbose:
        print("auto open command:", self.cmd_line)
      exit_code = pre_cmd.run(self.blkdev, self.rdisk)
      if self.args.verbose:
        print("auto open exit_code:", exit_code)
      if exit_code != 0:
        return exit_code
      self.blkdev = pre_cmd.blkdev
      # setup rdisk (if necessary)
      if cmd.need_rdisk():
        if not self._open_rdisk():
          raise IOError("No RDB Disk?")

    # run first command
    if self.args.verbose:
      print("command:", self.cmd_line)
    if cmd.edit and self.args.read_only:
      raise IOError("Edit commands not allowed in read-only mode")

    # check code of command after __init__ parsing
    if cmd.exit_code != 0:
      return cmd.exit_code

    # perform command
    exit_code = cmd.run(self.blkdev, self.rdisk)
    if cmd.blkdev != None:
      self.blkdev = cmd.blkdev
    if cmd.rdisk != None:
      self.rdisk = cmd.rdisk

    # final exit code
    if self.args.verbose:
      print("exit_code:", exit_code)
    return exit_code

  def run_next(self, cmd_line, cmd):
    self.cmd_line = cmd_line
    if self.args.verbose:
      print("command:", self.cmd_line)
    # verify command
    if cmd.edit and self.args.read_only:
      raise IOError("Edit commands not allowed in read-only mode")
    # make sure rdisk is set up
    if self.rdisk == None and cmd.need_rdisk():
      if not self._open_rdisk():
        raise IOError("No RDB Disk?")
    # run command
    exit_code = cmd.run(self.blkdev, self.rdisk)
    if cmd.blkdev != None:
      self.blkdev = cmd.blkdev
    if cmd.rdisk != None:
      self.rdisk = cmd.rdisk
    if self.args.verbose:
      print("exit_code:", exit_code)
    return exit_code

# ----- Commands -------------------------------------------------------------

# --- Open RDISK device/image ---

class OpenCommand(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts)
  def init_blkdev(self, file_name):
    # make sure image file exists
    if not os.path.exists(file_name):
      raise IOError("Image File not found: '%s'" % file_name)
    # open existing raw block device
    blkdev = RawBlockDevice(file_name, self.args.read_only)
    blkdev.open()
    # try to guess geometry
    geo = DiskGeometry()
    num_blocks = blkdev.num_blocks
    block_bytes = blkdev.block_bytes
    opts = KeyValue.parse_key_value_strings(self.opts)
    if geo.detect(num_blocks * block_bytes, opts) == None:
      raise IOError("Can't detect geometry of disk: '%s'" % file_name)
    blkdev.geo = geo
    return blkdev

# --- Create new RDISK device/image ---

class CreateCommand(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)
  def init_blkdev(self, file_name):
    # do not overwrite an existing image file
    if os.path.exists(file_name) and not self.args.force:
      raise IOError("Image File already exists: '%s'" % file_name)
    # make sure size is given
    if len(self.opts) < 1:
      print("Usage: create ( size=<n> | chs=<c,h,s> )")
      return None
    # determine disk geometry
    opts = KeyValue.parse_key_value_strings(self.opts)
    geo = DiskGeometry()
    if geo.setup(opts) == None:
      raise IOError("Can't set geometry of disk: '%s'" % file_name)
    else:
      # create new empty image file for geometry
      blkdev = RawBlockDevice(file_name)
      blkdev.create(geo.get_num_blocks())
      blkdev.geo = geo
      return blkdev

# --- Init existing disk image ---

class InitCommand(OpenCommand):
  def init_rdisk(self, blkdev):
    opts = KeyValue.parse_key_value_strings(self.opts)
    # number of cylinders for RDB
    if opts.has_key('rdb_cyls'):
      rdb_cyls = int(opts['rdb_cyls'])
    else:
      rdb_cyls = 1
    rdisk = RDisk(blkdev)
    rdisk.create(blkdev.geo, rdb_cyls=rdb_cyls)
    return rdisk

# --- Info about rdisk ----

class InfoCommand(Command):
  def handle_rdisk(self, rdisk):
    lines = rdisk.get_info()
    for l in lines:
      print(l)
    return 0

# --- Show rdisk structures ---

class ShowCommand(Command):
  def handle_rdisk(self, rdisk):
    show_hex = "hex" in self.opts
    rdisk.dump(show_hex)
    return 0

# --- Show allocation map ---

class MapCommand(Command):
  def handle_rdisk(self, rdisk):
    bm = rdisk.get_block_map()
    num = 0
    off = 0
    for i in bm:
      if num == 0:
        print("%06d: " % off, end="")
      print(i, end="")
      off += 1
      num += 1
      if num == 16:
        num = 0
        print("")
    return 0

# --- Free Partition Ranges

class FreeCommand(Command):
  def handle_rdisk(self, rdisk):
    ranges = rdisk.get_free_cyl_ranges()
    for r in ranges:
      print(r)
    return 0

# --- Add a partition ---

class PartEditCommand(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)

  def parse_opts(self, rdisk):
    self.popts = KeyValue.parse_key_value_strings(self.opts)
    self.rdisk = rdisk

  def get_dos_type(self, empty=False):
    if self.popts.has_key('fs'):
      fs_str = self.popts['fs']
    elif self.popts.has_key('dostype'):
      fs_str = self.popts['dostype']
    elif not empty:
      fs_str = self.args.dostype
    else:
      return None
    return parse_dos_type_str(str(fs_str))

  def get_drv_name(self, empty=False):
    if self.popts.has_key('name'):
      drv_name = self.popts['name']
    elif empty:
      drv_name = None
    else:
      drv_name = "%s%d" % (self.args.drive_prefix, self.rdisk.get_num_partitions())
    return drv_name

  def get_bootable(self, empty=False):
    if self.popts.has_key('bootable'):
      return bool(self.popts['bootable'])
    elif not empty:
      return False
    else:
      return None

  def get_boot_pri(self, empty=False):
    if self.popts.has_key('pri'):
      return self.popts['pri']
    elif not empty:
      return 0
    else:
      return None

  def get_automount(self, empty=False):
    if self.popts.has_key('automount'):
      return bool(self.popts['automount'])
    elif not empty:
      return True
    else:
      return None

  def get_flags(self, empty=False, old_flags=0):
    flags = 0
    bootable = self.get_bootable(empty=empty)
    if bootable is not None:
      if bootable:
        flags |= PartitionBlock.FLAG_BOOTABLE
    else:
      flags |= (old_flags) & PartitionBlock.FLAG_BOOTABLE
    automount = self.get_automount(empty=empty)
    if automount is not None:
      if not automount:
        flags |= PartitionBlock.FLAG_NO_AUTOMOUNT
    else:
      flags |= (old_flags) & PartitionBlock.FLAG_NO_AUTOMOUNT
    return flags

  def get_more_dos_env(self):
    more_dos_env = []
    valid_keys = PartitionDosEnv.valid_keys
    for key in self.popts:
      if key in valid_keys:
        more_dos_env.append((key, self.popts[key]))
    if len(more_dos_env) > 0:
      return more_dos_env
    else:
      return None

  def get_more_dos_env_info(self):
    valid_keys = PartitionDosEnv.valid_keys
    info = map(lambda x : "[%s=<n>]" % x, valid_keys)
    return " ".join(info)

  def get_cyl_range(self):
    start = None
    if self.popts.has_key('start'):
      start = int(self.popts['start'])
    # range with start=<n> end=<n>
    if self.popts.has_key('end'):
      end = int(self.popts['end'])
      if start == None or end <= start:
        return None
      else:
        return (start, end)
    # expect a size
    elif self.popts.has_key('size'):
      size = self.popts['size']
      cyls = None
      if type(size) == int:
        cyls = size
      # size in bytes
      elif size[-1] in ('b','B'):
        bytes = ByteSize.parse_byte_size_str(size[:-1])
        if bytes == None:
          return None
        cyls = bytes / self.rdisk.get_cylinder_bytes()
      # size in percent
      elif size[-1] == '%':
        prc = float(size[:-1])
        cyls = int(prc * self.rdisk.get_logical_cylinders() / 100.0)
      # size in cylinders
      else:
        cyls = ByteSize.parse_byte_size_str(size)

      # check cyls
      if cyls == None or cyls < 1:
        return None
      # find a range if no start is given
      if start == None:
        start = self.rdisk.find_free_cyl_range_start(cyls)
        if start == None:
          return None
      return (start, start + cyls - 1)
    # nothing specified -> get next free range
    else:
      ranges = self.rdisk.get_free_cyl_ranges()
      if ranges == None:
        return None
      return ranges[0]

class AddCommand(PartEditCommand):
  def handle_rdisk(self, rdisk):
    self.parse_opts(rdisk)
    lo_hi = self.get_cyl_range()
    if lo_hi == None:
      print("ERROR: invalid partition range given!")
      return 1
    dostype = self.get_dos_type()
    if dostype == None:
      print("ERROR: invalid dos type!")
      return 1
    drv_name = self.get_drv_name()
    if drv_name == None:
      print("ERROR: invalid drive name!")
    flags = self.get_flags()
    boot_pri = self.get_boot_pri()
    more_dos_env = self.get_more_dos_env()
    print("creating: '%s' %s %s" % (drv_name, lo_hi, num_to_tag_str(dostype)))
    # add partition
    if rdisk.add_partition(drv_name, lo_hi, dos_type=dostype, flags=flags, boot_pri=boot_pri, more_dos_env=more_dos_env):
      return 0
    else:
      print("ERROR: creating partition: '%s': %s" % (drv_name, lo_hi))
      return 1

class ChangeCommand(PartEditCommand):
  def handle_rdisk(self, rdisk):
    if len(self.opts) < 1:
      print("Usage: change <id> [name=<s>] [dostype=<n|tag>] [automount=<b>] [bootable=<b>] [pri=<n>] " + self.get_more_dos_env_info())
      return 1
    else:
      p = rdisk.find_partition_by_string(self.opts[0])
      if p != None:
        self.parse_opts(rdisk)
        dostype = self.get_dos_type(empty=True)
        drv_name = self.get_drv_name(empty=True)
        flags = self.get_flags(empty=True, old_flags=p.get_flags())
        boot_pri = self.get_boot_pri(empty=True)
        more_dos_env = self.get_more_dos_env()
        # change partition
        if rdisk.change_partition(p.num, drv_name=drv_name, dos_type=dostype, flags=flags, boot_pri=boot_pri, more_dos_env=more_dos_env):
          return 0
        else:
          print("ERROR: changing partition: '%s'" % (drv_name))
          return 1
      else:
        print("Can't find partition: '%s'" % self.opts[0])
        return 1

# --- Fill empty space with partitions ---

class FillCommand(PartEditCommand):
  def handle_rdisk(self, rdisk):
    self.parse_opts(rdisk)
    ranges = rdisk.get_free_cyl_ranges()
    # nothing to do
    if ranges == None:
      return 0
    for lo_hi in ranges:
      drv_name = self.get_drv_name()
      if drv_name == None:
        print("ERROR: invalid drive name!")
      dostype = self.get_dos_type()
      if dostype == None:
        print("ERROR: invalid dostype given!")
        return 1
      print("creating: '%s' %s %s" % (drv_name, lo_hi, num_to_tag_str(dostype)))
      # add partition
      if not rdisk.add_partition(drv_name, lo_hi, dos_type=dostype):
        print("ERROR: creating partition: '%s': %s" % (drv_name, lo_hi))
        return 1
    return 0

# --- Delete partition command ---

class DeleteCommand(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)

  def handle_rdisk(self, rdisk):
    if len(self.opts) < 1:
      print("Usage: delete <id>")
      return 1
    else:
      p = rdisk.find_partition_by_string(self.opts[0])
      if p != None:
        if not rdisk.delete_partition(p.num):
          print("ERROR: deleting partition: '%s'" % self.opts[0])
          return 1
        else:
          return 0
      else:
        print("Can't find partition: '%s'" % self.opts[0])
        return 1

# --- Filesystem Commands ---

class FSGetCommand(Command):
  def handle_rdisk(self, rdisk):
    if len(self.opts) < 2:
      print("Usage: fsget <id> <file_name>")
      return 1
    else:
      num = int(self.opts[0])
      fs = rdisk.get_filesystem(num)
      if fs == None:
        print("fsget: invalid filesystem index",num)
        return 1
      else:
        file_name = self.opts[1]
        data = fs.get_data()
        f = open(file_name,"wb")
        f.write(data)
        f.close()
        return 0

class FSAddCommand(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)

  def parse_opts(self):
    self.popts = KeyValue.parse_key_value_strings(self.opts)

  def get_dos_type(self):
    if self.popts.has_key('fs'):
      fs_str = self.popts['fs']
    elif self.popts.has_key('dostype'):
      fs_str = self.popts['dostype']
    else:
      fs_str = self.args.dostype
    return parse_dos_type_str(str(fs_str))

  def handle_rdisk(self, rdisk):
    self.parse_opts()
    valid_flags = FSHeaderDeviceNode.valid_flags
    if len(self.opts) < 1:
      flag_info = map(lambda x : "[%s=<n>]" % x, valid_flags)
      flag_info = " ".join(flag_info)
      print("Usage: fsadd <file_name> [dostype=<n|tag>] [version=<n.m>] " + flag_info)
      return 1
    else:
      # parse options
      opts = KeyValue.parse_key_value_strings(self.opts)
      # read file data
      file_name = self.opts[0]
      f = open(file_name,"rb")
      data = f.read()
      f.close()
      # get version from binary
      tag = VerTag.find(data)
      ver = None
      if tag != None:
        ver = VerTag.get_version(tag)
      if ver == None:
        ver = (0,0)
      # overwrite version from options
      if opts.has_key('version'):
        vstr = opts['version']
        pos = vstr.find('.')
        if pos != -1:
          ver = (int(vstr[:pos]),int(vstr[pos+1:]))
      # valid fs flags
      dev_flags = []
      for key in opts:
        if key in valid_flags:
          dev_flags.append((key,opts[key]))
      # add fs
      version = ver[0] << 16 | ver[1]
      # get dostype
      dostype = self.get_dos_type()
      if rdisk.add_filesystem(data, dos_type=dostype, version=version, dev_flags=dev_flags):
        return 0
      else:
        print("ERROR adding filesystem! (no space in RDB left)")
        return 1

class FSDeleteCommand(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)

  def handle_rdisk(self, rdisk):
    if len(self.opts) < 1:
      print("Usage: fsdelete <fid>")
      return 1
    else:
      fs = rdisk.find_filesystem_by_string(self.opts[0])
      if fs != None:
        if not rdisk.delete_filesystem(fs.num):
          print("ERROR deleting filesystem: '%s'" % self.opts[0])
          return 1
        else:
          return 0
      else:
        print("ERROR finding filesystem: '%s'" % self.opts[0])
        return 1

class FSFlagsCommand(Command):
  def __init__(self, args, opts):
    Command.__init__(self, args, opts, edit=True)

  def handle_rdisk(self, rdisk):
    if len(self.opts) < 2:
      print("Usage: fsflags <fid> [ clear | key=<val> ... ]")
      return 1
    else:
      fs = rdisk.find_filesystem_by_string(self.opts[0])
      if fs != None:
        opts = KeyValue.parse_key_value_strings(self.opts[1:])
        valid_flags = fs.get_valid_flag_names()
        flags = []
        clear = False
        for o in opts:
          if o in valid_flags:
            flags.append((o,opts[o]))
          elif o == 'clear':
            clear = True
        fs.set_flags(flags, clear)
        return 0
      else:
        print("ERROR finding filesystem: '%s'" % self.opts[0])
        return 1

# ----- main -----
def main():
  # call scanner and process all files with selected command
  cmd_map = {
  "open" : OpenCommand,
  "create" : CreateCommand,
  "init" : InitCommand,
  "info" : InfoCommand,
  "show" : ShowCommand,
  "free" : FreeCommand,
  "add" : AddCommand,
  "fill" : FillCommand,
  "fsget" : FSGetCommand,
  "fsadd" : FSAddCommand,
  "fsdelete" : FSDeleteCommand,
  "fsflags" : FSFlagsCommand,
  "map" : MapCommand,
  "delete" : DeleteCommand,
  "change" : ChangeCommand
  }

  parser = argparse.ArgumentParser()
  parser.add_argument('image_file')
  parser.add_argument('command_list', nargs='+', help="command: "+",".join(cmd_map.keys()))
  parser.add_argument('-v', '--verbose', action='store_true', default=False, help="be more verbos")
  parser.add_argument('-s', '--seperator', default='+', help="set the command separator char sequence")
  parser.add_argument('-r', '--read-only', action='store_true', default=False, help="read-only operation")
  parser.add_argument('-f', '--force', action='store_true', default=False, help="force overwrite existing image")
  parser.add_argument('-p', '--drive-prefix', default='DH', help="set default drive name prefix (DH -> DH0, DH1, ...)")
  parser.add_argument('-t', '--dostype', default='ffs+intl', help="set default dos type")
  args = parser.parse_args()

  cmd_list = args.command_list
  sep = args.seperator
  queue = FSCommandQueue(args, cmd_list, sep, cmd_map)
  code = queue.run()
  return code


if __name__ == '__main__':
  sys.exit(main())



import time


class Notification(object):
    DEFAULT_DURATION = 5.0
    notifications = []
    new = False

    def __init__(self, text, duration=DEFAULT_DURATION):
        self.start = 0.0
        self.end = 0.0
        self.duration = duration
        self.text = text
        self.add(self)

    def show(self):
        if self.start > 0:
            return
        self.start = time.time()
        self.end = self.start + self.duration

    def dismiss(self):
        self.start = 1
        self.end = 1

    @classmethod
    def add(cls, notification):
        cls.notifications.append(notification)
        cls.new = True
        print("\n" + "-" * 79 + "\n" + "NOTIFICATION")
        print(notification.text)

    @classmethod
    def all(cls):
        notifications = []
        for notification in cls.notifications:
            if notification.end == 0 or notification.end > time.time():
                notifications.append(notification)
        if len(cls.notifications) != len(notifications):
            cls.notifications = notifications[:]
        return notifications

    @classmethod
    def count(cls):
        return len(cls.notifications)



from __future__ import absolute_import
from __future__ import print_function

from .BlockDevice import BlockDevice
import ctypes
import gzip
import io

class ADFBlockDevice(BlockDevice):
  def __init__(self, adf_file, read_only=False, fobj=None):
    self.adf_file = adf_file
    self.read_only = read_only
    self.fobj = fobj
    self.dirty = False
    lo = adf_file.lower()
    self.gzipped = lo.endswith('.adz') or lo.endswith('.adf.gz')

  def create(self):
    if self.read_only:
      raise IOError("ADF creation not allowed in read-only mode!")
    self._set_geometry() # set default geometry
    # allocate image in memory
    self.data = ctypes.create_string_buffer(self.num_bytes)
    self.dirty = True

  def open(self):
    self._set_geometry() # set default geometry
    close = True
    # open adf file via fobj
    if self.fobj is not None:
      if self.gzipped:
        fh = gzip.GzipFile(self.adf_file, "rb", fileobj=self.fobj)
      else:
        fh = self.fobj
        close = False
    # open adf file
    else:
      if self.gzipped:
        fh = gzip.open(self.adf_file,"rb")
      else:
        fh = io.open(self.adf_file, "rb")
    # read image
    data = fh.read(self.num_bytes)
    # close input file
    if close:
      fh.close()
    # check size
    if len(data) != self.num_bytes:
      raise IOError("Invalid ADF Size: got %d but expected %d" % (len(data), self.num_bytes))
    # create modifyable data
    if self.read_only:
      self.data = data
    else:
      self.data = ctypes.create_string_buffer(data)

  def flush(self):
    # write dirty adf
    if self.dirty and not self.read_only:
      close = True
      if self.fobj is not None:
        # seek fobj to beginning
        self.fobj.seek(0,0)
        if self.gzipped:
          fh = gzip.GzipFile(self.adf_file, "wb", fileobj=self.fobj)
        else:
          fh = self.fobj
          close = False
      else:
        if self.gzipped:
          fh = gzip.open(self.adf_file,"wb")
        else:
          fh = io.open(self.adf_file, "wb")
      # write image
      fh.write(self.data)
      # close file
      if close:
        fh.close()
      self.dirty = False

  def close(self):
    self.flush()
    self.data = None
    # now close fobj
    if self.fobj is not None:
      self.fobj.close()

  def read_block(self, blk_num):
    if blk_num >= self.num_blocks:
      raise ValueError("Invalid ADF block num: got %d but max is %d" % (blk_num, self.num_blocks))
    off = self._blk_to_offset(blk_num)
    return self.data[off:off+self.block_bytes]

  def write_block(self, blk_num, data):
    if self.read_only:
      raise IOError("ADF File is read-only!")
    if blk_num >= self.num_blocks:
      raise ValueError("Invalid ADF block num: got %d but max is %d" % (blk_num, self.num_blocks))
    if len(data) != self.block_bytes:
      raise ValueError("Invalid ADF block size written: got %d but size is %d" % (len(data), self.block_bytes))
    off = self._blk_to_offset(blk_num)
    self.data[off:off+self.block_bytes] = data
    self.dirty = True


# --- mini test ---
if __name__ == '__main__':
  import sys
  for a in sys.argv[1:]:
    # write to file device
    adf = ADFBlockDevice(a)
    adf.open()
    d = adf.read_block(0)
    adf.write_block(0, d)
    adf.close()
    # write via fobj
    fobj = open(a, "rb")
    adf = ADFBlockDevice(a, fobj=fobj)
    adf.open()
    d = adf.read_block(0)
    adf.write_block(0, d)
    adf.close()



import os
import fsui
import workspace.path


class Image(fsui.Image):
    def __init__(self, path):
        if hasattr(path, "read"):
            # Stream
            pass
        elif os.path.isfile(path):
            # Native path
            pass
        else:
            path = workspace.path.host(path)
        super().__init__(path)



from __future__ import absolute_import
from __future__ import print_function

from amitools.fs.validate.Log import Log
import struct

class BitmapScan:
  """Validate the bitmap of a file system"""
  
  def __init__(self, block_scan, log):
    self.block_scan = block_scan
    self.log = log
    self.bm_blocks = None
    
  def scan_bitmap(self, root):
    """scan the file system bitmap"""
    # first check bitmap flag
    bm_flag = root.bitmap_flag
    if bm_flag != 0xffffffff:
      self.log.msg(Log.WARN,"Root bitmap flag not valid (-1)",root.blk_num)
    # now calculate the size of the bitmap
    num_blks = self.block_scan.blkdev.num_blocks - self.block_scan.blkdev.reserved
    block_longs = self.block_scan.blkdev.block_longs - 1 # all longs are available for bitmap
    self.num_bm_lwords = int((num_blks + 31) / 32) # 32 blocks fit in a long word
    self.num_bm_blocks = int((self.num_bm_lwords + block_longs - 1) / block_longs)
    self.log.msg(Log.DEBUG,"Total Bitmap DWORDs: %d  (block %d)" % (self.num_bm_lwords, block_longs))
    self.log.msg(Log.DEBUG,"Number of Bitmap Blocks: %d" % self.num_bm_blocks)
    # calc the bitmask in the last word
    last_filled_bits = self.num_bm_lwords * 32 - num_blks
    if last_filled_bits == 32:
      self.last_mask = 0xffffffff
    else:
      self.last_mask = (1 << last_filled_bits) - 1
    self.log.msg(Log.DEBUG,"Last DWORD mask: %08x" % self.last_mask)    
    # now scan bitmap blocks and build list of all bitmap blocks
    self.read_bitmap_ptrs_and_blocks(root)
    found_blocks = len(self.bm_blocks)
    self.log.msg(Log.DEBUG,"Found Bitmap Blocks: %d" % found_blocks)
    # check number of blocks
    if found_blocks != self.num_bm_blocks:
      self.log.msg(Log.ERROR,"Invalid number of Bitmap Blocks: found=%d expected=%d" % (found_blocks, self.num_bm_blocks), root.blk_num)
    else:
      # check bits in bitmap
      self.check_bits()
    
  def check_bits(self):
    """calculate allocation bits and verify with stored ones"""
    # block range
    blkdev = self.block_scan.blkdev
    # first bitmap data
    cur_pos = 0
    bm_blk = 0
    cur_data = self.bm_blocks[0].bitmap
    blk_size = len(cur_data)
    # loop throug all bitmap longwords
    lw = 0
    blk_num = blkdev.reserved
    max_lw = self.num_bm_lwords - 1
    while lw < max_lw:
      got = struct.unpack_from(">I",cur_data,cur_pos)[0]
      expect = self.calc_lword(blk_num)
      if got != expect:
        self.log.msg(Log.ERROR,"Invalid bitmap allocation (@%d: #%d+%d) blks [%d...%d] got=%08x expect=%08x" \
          % (lw, bm_blk, cur_pos/4, blk_num, blk_num+31, got, expect))
      lw += 1
      blk_num += 32
      # fetch next bitmap data block
      cur_pos += 4
      if cur_pos == blk_size:
        bm_blk += 1
        cur_data = self.bm_blocks[bm_blk].bitmap
        cur_pos = 0
    # the last long word
    got = struct.unpack_from(">I",cur_data,cur_pos)[0] & self.last_mask
    expect = self.calc_lword(blk_num) & self.last_mask
    if got != expect:
      self.log.msg(Log.ERROR,"Invalid bitmap allocation (last) (@%d: #%d+%d) blks [%d...%d] got=%08x expect=%08x" \
        % (lw, bm_blk, cur_pos/4, blk_num, blkdev.num_blocks-1, got, expect))
    
  def calc_lword(self, blk_num):
    """calcuate the bitmap lword"""
    value = 0
    for i in range(32):
      # set bit in lword if block is available
      if not self.block_scan.is_block_available(blk_num + i):
        mask = 1 << i
        value |= mask
    return value
    
  def read_bitmap_ptrs_and_blocks(self, root):
    """build the list of all file system bitmap blocks"""
    self.bm_blocks = []
    # scan list embedded in root block
    self.read_bm_list(root.bitmap_ptrs, root.blk_num)
    # now follow bitmap extension blocks
    cur_blk_num = root.blk_num
    bm_ext = root.bitmap_ext_blk
    while bm_ext != 0:
      # check ext block
      if self.block_scan.is_block_available(bm_ext):
        self.log.msg(Log.ERROR,"Bitmap ext block @%d already used?" % bm_block, cur_blk_num)
      else:
        bi = self.block_scan.read_block(bm_ext, is_bm_ext=True)
        if bi == None:
          self.log.msg(Log.ERROR,"Error reading bitmap ext block @%d" % bm_ext, cur_blk_num)
          break
        else:
          self.read_bm_list(bi.bitmap_ptrs, bm_ext)
          cur_blk_num = bm_ext
          bm_ext = bi.next_blk
    
  def read_bm_list(self, ptrs, blk_num):
    list_end = False
    for bm_block in ptrs:
      # still inside the pointer list
      if list_end == False:
        # add a normal block
        if bm_block != 0:
          # make sure bitmap block was not used already
          if self.block_scan.is_block_available(bm_block):
            self.log.msg(Log.ERROR,"Bitmap block @%d already used?" % bm_block, blk_num)
          else:
            # read bitmap block
            bi = self.block_scan.read_block(bm_block, is_bm=True)
            if bi == None:
              self.log.msg(Log.ERROR,"Error reading bitmap block @%d" % bm_block, blk_num)
            else:
              self.bm_blocks.append(bi)
        else:
          list_end = True
      else:
        # make sure no blocks are referenced
        if bm_block != 0:
          self.log.msg(Log.ERROR,"Referenced bitmap block @%d beyond end of list" % bm_block, blk_num)



from functools import lru_cache

from fsgs.drivers.gamedriver import GameDriver


class PlatformHandler:
    def __init__(self, loader_class=None, runner_class=None):
        self.loader_class = loader_class
        self.runner_class = runner_class

    @classmethod
    def create(cls, platform_id):
        class_ = cls.get_platform_class(platform_id)
        return class_()

    @classmethod
    def get_platform_class(cls, platform_id):
        platform_id = platform_id.lower()
        try:
            return platforms[platform_id]
        except KeyError:
            return UnsupportedPlatform

    @classmethod
    @lru_cache()
    def get_platform_name(cls, platform_id):
        return cls.get_platform_class(platform_id).PLATFORM_NAME

    @classmethod
    @lru_cache()
    def get_platform_ids(cls):
        return sorted(platforms.keys())

    def get_loader(self, fsgs):
        if self.loader_class is None:
            raise Exception("loader class is None for " + repr(self))
        return self.loader_class(fsgs)

    def get_runner(self, fsgs) -> GameDriver:
        if self.runner_class is None:
            raise Exception("runner class is None for " + repr(self))
        return self.runner_class(fsgs)

    def driver(self, fsgc):
        return self.get_runner(fsgc)

    def loader(self, fsgc):
        return self.get_loader(fsgc)


class Platform(PlatformHandler):
    AMIGA = "amiga"
    ARCADE = "arcade"
    A2600 = "a2600"
    A5200 = "a5200"
    A7800 = "a7800"
    ATARI = "atari"
    C64 = "c64"
    CD32 = "cd32"
    CDTV = "cdtv"
    CPC = "cpc"
    DOS = "dos"
    GB = "gb"
    GBA = "gba"
    GBC = "gbc"
    LYNX = "lynx"
    MSX = "msx"
    N64 = "n64"
    NDS = "nds"
    NEOGEO = "neogeo"
    NES = "nes"
    NGC = "ngc"
    PSX = "psx"
    SNES = "snes"
    SGG = "sgg"
    SMD = "smd"
    SMS = "sms"
    TG16 = "tg16"
    TGCD = "tgcd"
    ZXS = "zxs"

    def get_loader(self, fsgc):
        return self.loader(fsgc)

    def get_runner(self, fsgc):
        return self.driver(fsgc)

    def driver(self, fsgc):
        raise NotImplementedError()

    def loader(self, fsgc):
        raise NotImplementedError()


from fsgs.platforms.amiga import AmigaPlatformHandler  # noqa: E402
from fsgs.platforms.amstrad_cpc import AmstradCPCPlatformHandler  # noqa: E402
from fsgs.platforms.arcade.arcadeplatform import ArcadePlatformHandler  # noqa: E402
from fsgs.platforms.atari5200 import Atari5200PlatformHandler  # noqa: E402
from fsgs.platforms.atari7800 import Atari7800PlatformHandler  # noqa: E402
from fsgs.platforms.atari.atariplatform import AtariSTPlatformHandler  # noqa: E402
from fsgs.platforms.atari_2600 import Atari2600PlatformHandler  # noqa: E402
from fsgs.platforms.cd32 import CD32PlatformHandler  # noqa: E402
from fsgs.platforms.cdtv import CDTVPlatformHandler  # noqa: E402
from fsgs.platforms.commodore64 import Commodore64Platform  # noqa: E402
from fsgs.platforms.dos.dosplatform import DOSPlatformHandler  # noqa: E402
from fsgs.platforms.gameboy import GameBoyPlatform  # noqa: E402
from fsgs.platforms.gameboyadvance import GameBoyAdvancePlatform  # noqa: E402
from fsgs.platforms.gameboycolor import GameBoyColorPlatform  # noqa: E402
from fsgs.platforms.gamecube import GameCubePlatform  # noqa: E402
from fsgs.platforms.gamegear import GameGearPlatform  # noqa: E402
from fsgs.platforms.lynx import LynxPlatformHandler  # noqa: E402
from fsgs.platforms.mastersystem import MasterSystemPlatform  # noqa: E402
from fsgs.platforms.megadrive import MegaDrivePlatform  # noqa: E402
from fsgs.platforms.msx import MsxPlatformHandler  # noqa: E402
from fsgs.platforms.neogeo import NeoGeoPlatform  # noqa: E402
from fsgs.platforms.nintendo import NintendoPlatform  # noqa: E402
from fsgs.platforms.nintendo64 import Nintendo64Platform  # noqa: E402
from fsgs.platforms.nintendods import NintendoDSPlatform  # noqa: E402
from fsgs.platforms.playstation import PlayStationPlatform  # noqa: E402
from fsgs.platforms.supernintendo import SuperNintendoPlatformHandler  # noqa: E402
from fsgs.platforms.turbografx16 import TurboGrafx16Platform  # noqa: E402
from fsgs.platforms.turbografxcd import TurboGrafxCDPlatform  # noqa: E402
from fsgs.platforms.zxs import SpectrumPlatformHandler  # noqa: E402


class UnsupportedPlatform(PlatformHandler):
    PLATFORM_NAME = "Unsupported"


platforms = {
    Platform.AMIGA: AmigaPlatformHandler,
    Platform.ARCADE: ArcadePlatformHandler,
    Platform.A2600: Atari2600PlatformHandler,
    Platform.A5200: Atari5200PlatformHandler,
    Platform.A7800: Atari7800PlatformHandler,
    Platform.ATARI: AtariSTPlatformHandler,
    Platform.C64: Commodore64Platform,
    Platform.CD32: CD32PlatformHandler,
    Platform.CDTV: CDTVPlatformHandler,
    Platform.CPC: AmstradCPCPlatformHandler,
    Platform.DOS: DOSPlatformHandler,
    Platform.GB: GameBoyPlatform,
    Platform.GBA: GameBoyAdvancePlatform,
    Platform.GBC: GameBoyColorPlatform,
    Platform.LYNX: LynxPlatformHandler,
    Platform.MSX: MsxPlatformHandler,
    Platform.N64: Nintendo64Platform,
    Platform.NDS: NintendoDSPlatform,
    Platform.NEOGEO: NeoGeoPlatform,
    Platform.NES: NintendoPlatform,
    Platform.NGC: GameCubePlatform,
    Platform.SNES: SuperNintendoPlatformHandler,
    Platform.PSX: PlayStationPlatform,
    Platform.SGG: GameGearPlatform,
    Platform.SMD: MegaDrivePlatform,
    Platform.SMS: MasterSystemPlatform,
    Platform.TG16: TurboGrafx16Platform,
    Platform.TGCD: TurboGrafxCDPlatform,
    Platform.ZXS: SpectrumPlatformHandler,
}
PLATFORM_IDS = platforms.keys()


def normalize_platform_id(platform_id):
    platform_id = platform_id.lower().replace("-", "").replace("_", "")
    # noinspection SpellCheckingInspection
    if platform_id in ["st", "atarist"]:
        return Platform.ATARI
    elif platform_id in ["commodorecdtv"]:
        return Platform.CDTV
    elif platform_id in ["amigacd32"]:
        return Platform.CD32
    elif platform_id in ["amstradcpc"]:
        return Platform.CPC
    elif platform_id in ["msdos"]:
        return Platform.DOS
    elif platform_id in ["gameboy"]:
        return Platform.GB
    elif platform_id in ["gameboyadvance"]:
        return Platform.GBA
    elif platform_id in ["gameboycolor"]:
        return Platform.GBC
    elif platform_id in ["nintendo", "famicom"]:
        return Platform.NES
    elif platform_id in ["nintendo64"]:
        return Platform.N64
    elif platform_id in ["nintendods"]:
        return Platform.NDS
    elif platform_id in ["nintendogamecube", "gamecube", "gc"]:
        return Platform.NGC
    elif platform_id in ["supernintendo", "supernes", "superfamicom"]:
        return Platform.SNES
    elif platform_id in ["zxspectrum"]:
        return Platform.ZXS
    elif platform_id in ["mastersystem", "segamastersystem"]:
        return Platform.SMS
    elif platform_id in ["megadrive", "segamegadrive", "genesis"]:
        return Platform.SMD
    elif platform_id in ["atari2600"]:
        return Platform.A2600
    elif platform_id in ["atari5200"]:
        return Platform.A5200
    elif platform_id in ["atari78600"]:
        return Platform.A7800
    elif platform_id in ["turbografx16"]:
        return Platform.TG16
    elif platform_id in ["gamegear", "sgg"]:
        return Platform.SGG
    return platform_id



import json
import os
import sqlite3
import zlib
from binascii import hexlify, unhexlify

from .BaseDatabase import BaseDatabase

VERSION = 19
RESET_VERSION = 19
DUMMY_UUID = b"'\\x8b\\xbb\\x00Y\\x8bqM\\x15\\x972\\xa8-t_\\xb2\\xfd'"


class IncompleteGameException(Exception):
    pass


class GameDatabase(BaseDatabase):
    def __init__(self, path):
        BaseDatabase.__init__(self, BaseDatabase.SENTINEL)
        self._path = path
        self._name = os.path.splitext(os.path.basename(path))[0]

    def name(self):
        return self._name

    def path(self):
        return self._path

    def get_path(self):
        return self._path

    def version(self):
        return VERSION

    def get_version(self):
        return VERSION

    def get_reset_version(self):
        return RESET_VERSION

    @staticmethod
    def query(q):
        return q.replace("%s", "?")

    def get_last_game_id(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT max(id) FROM game")
        row = cursor.fetchone()
        return row[0] or 0

    def get_ratings_for_game(self, game_uuid):
        cursor = self.internal_cursor()
        cursor.execute(
            "SELECT like_rating, work_rating FROM rating WHERE game_uuid = ?",
            (game_uuid,),
        )
        row = cursor.fetchone()
        if row is not None:
            return row[0], row[1]
        else:
            return 0, 0

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def get_license_code_for_url(self, url):
        # cursor = self.internal_cursor()
        # cursor.execute(
        #     "SELECT license_code FROM file "
        #     "WHERE external = ? LIMIT 1", (url,))
        # row = cursor.fetchone()
        # if not row:
        #     return None
        # return row[0]
        return None

    def add_game(self, game_id, game_uuid, game_data):
        cursor = self.internal_cursor()
        cursor.execute(
            "DELETE FROM game WHERE uuid = ?", (sqlite3.Binary(game_uuid),)
        )
        cursor.execute(
            "INSERT INTO game (id, uuid, data) VALUES (?, ?, ?)",
            (game_id, sqlite3.Binary(game_uuid), sqlite3.Binary(game_data)),
        )

    def delete_game(self, game_id, game_uuid):
        cursor = self.internal_cursor()
        cursor.execute(
            "DELETE FROM game WHERE uuid = ?", (sqlite3.Binary(game_uuid),)
        )
        cursor.execute(
            "DELETE FROM game WHERE uuid = ?", (sqlite3.Binary(DUMMY_UUID),)
        )
        cursor.execute(
            "INSERT INTO game (id, uuid, data) VALUES (?, ?, ?)",
            (game_id, sqlite3.Binary(DUMMY_UUID), ""),
        )

    @staticmethod
    def binary_uuid_to_str(data):
        s = hexlify(data).decode("ASCII")
        return "{}-{}-{}-{}-{}".format(
            s[0:8], s[8:12], s[12:16], s[16:20], s[20:32]
        )

    def get_all_uuids(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT uuid FROM game WHERE data IS NOT NULL")
        result = set()
        for row in cursor:
            data = row[0]
            if len(data) != 16:
                print(
                    "WARNING: Invalid UUID ({} bytes) found: {}".format(
                        len(data), repr(data)
                    )
                )
                continue
            result.add(self.binary_uuid_to_str(data))
        return result

    def get_game_values_for_uuid(self, game_uuid, recursive=True):
        # print("get_game_values_for_uuid", game_uuid)
        assert game_uuid
        assert isinstance(game_uuid, str)
        return self.get_game_values(game_uuid=game_uuid, recursive=recursive)

    def get_game_values(self, game_id=None, *, game_uuid=None, recursive=True):
        cursor = self.internal_cursor()
        if game_uuid is not None:
            cursor.execute(
                "SELECT uuid, data FROM game WHERE uuid = ?",
                (sqlite3.Binary(unhexlify(game_uuid.replace("-", ""))),),
            )
            row = cursor.fetchone()
            if not row:
                raise LookupError("Cannot find game uuid {}".format(game_uuid))
        else:
            cursor.execute(
                "SELECT uuid, data FROM game WHERE id = ?", (game_id,)
            )
            row = cursor.fetchone()
            if not row:
                raise LookupError("Cannot find game id {}".format(game_id))
        if game_uuid is None:
            game_uuid = self.binary_uuid_to_str(row[0])
        else:
            assert self.binary_uuid_to_str(row[0]) == game_uuid
        data = zlib.decompress(row[1])
        data = data.decode("UTF-8")

        doc = json.loads(data)
        doc["variant_uuid"] = game_uuid

        # This is a hack so we can retrieve the published-status of the
        # child entry. The original key should have been named _published,
        # probably, to avoid it being inherited.
        variant_published = doc.get("publish", "")

        next_parent_uuid = doc.get("parent_uuid", "")
        next_parent_database = doc.get("parent_database", "")
        while next_parent_uuid and recursive:
            # Treat game_uuid special, it will be the first parent_uuid
            # in the chain.
            doc["game_uuid"] = next_parent_uuid
            if next_parent_database:
                break
            else:
                cursor.execute(
                    "SELECT data FROM game WHERE uuid = ?",
                    (
                        sqlite3.Binary(
                            unhexlify(next_parent_uuid.replace("-", ""))
                        ),
                    ),
                )
                row = cursor.fetchone()
                if not row:
                    raise IncompleteGameException(
                        "Could not find parent {0} of game {1}".format(
                            next_parent_uuid, game_uuid
                        )
                    )
                data = zlib.decompress(row[0])
                data = data.decode("UTF-8")
                next_doc = json.loads(data)
            next_parent_uuid = next_doc.get("parent_uuid", "")
            next_parent_database = next_doc.get("parent_database", "")
            # Let child doc overwrite and append values to parent doc.
            next_doc.update(doc)
            doc = next_doc

        doc["__publish_hack__"] = variant_published
        return doc

    def get_game_database_version(self):
        cursor = self.internal_cursor()
        cursor.execute("SELECT database_version FROM metadata")
        return cursor.fetchone()[0]

    def set_game_database_version(self, version):
        cursor = self.internal_cursor()
        cursor.execute("UPDATE metadata SET database_version = ?", (version,))

    def clear(self):
        cursor = self.internal_cursor()
        cursor.execute("DELETE FROM rating")
        cursor.execute("DELETE FROM game")

    def update_database_to_version_19(self):
        cursor = self.internal_cursor()
        cursor.execute(
            """
            ALTER TABLE metadata ADD COLUMN games_version
            INTEGER NOT NULL DEFAULT 0;
        """
        )
        cursor.execute(
            """
            CREATE TABLE game (
                id INTEGER PRIMARY KEY,
                uuid BLOB,
                data BLOB
            );
        """
        )
        cursor.execute(
            """
            CREATE INDEX game_uuid ON game (uuid);
        """
        )
        cursor.execute(
            """
            CREATE TABLE rating (
                game_uuid VARCHAR(36) PRIMARY KEY NOT NULL,
                work_rating INT NOT NULL,
                like_rating INT NOT NULL,
                updated TIMESTAMP NULL
            );
        """
        )
        cursor.execute(
            """
            ALTER TABLE metadata ADD COLUMN database_version
            INTEGER NOT NULL DEFAULT 0;
        """
        )



from __future__ import absolute_import
from __future__ import print_function

from .ADFSVolume import ADFSVolume
from amitools.fs.blkdev.BlkDevFactory import BlkDevFactory

class Repacker:
  def __init__(self, in_image_file, in_options=None):
    self.in_image_file = in_image_file
    self.in_options = in_options
    self.in_blkdev = None
    self.out_blkdev = None
    self.in_volume = None
    self.out_volume = None
  
  def create_in_blkdev(self):
    f = BlkDevFactory()
    self.in_blkdev = f.open(self.in_image_file, read_only=True, options=self.in_options)
    return self.in_blkdev  
      
  def create_in_volume(self):
    if self.in_blkdev == None:
      return None
    self.in_volume = ADFSVolume(self.in_blkdev)
    self.in_volume.open()
    return self.in_volume
  
  def create_in(self):
    if self.create_in_blkdev() == None:
      return False
    if self.create_in_volume() == None:
      return False
    return True
  
  def create_out_blkdev(self, image_file, force=True, options=None):
    if self.in_blkdev == None:
      return None
    # clone geo from input
    if options == None:
      options = self.in_blkdev.get_chs_dict()
    f = BlkDevFactory()
    self.out_blkdev = f.create(image_file, force=force, options=options)
    return self.out_blkdev
  
  def create_out_volume(self, blkdev=None):
    if blkdev != None:
      self.out_blkdev = blkdev
    if self.out_blkdev == None:
      return None
    if self.in_volume == None:
      return None
    # clone input volume
    iv = self.in_volume
    name = iv.get_volume_name()
    dos_type = iv.get_dos_type()
    meta_info = iv.get_meta_info()
    boot_code = iv.get_boot_code()
    self.out_volume = ADFSVolume(self.out_blkdev)
    self.out_volume.create(name, meta_info=meta_info, dos_type=dos_type, boot_code=boot_code)
    return self.out_volume
    
  def repack(self):
    self.repack_node_dir(self.in_volume.get_root_dir(), self.out_volume.get_root_dir())
  
  def repack_node_dir(self, in_root, out_root):
    entries = in_root.get_entries()
    for e in entries:
      self.repack_node(e, out_root) 
  
  def repack_node(self, in_node, out_dir):
    name = in_node.get_file_name().get_name()
    meta_info = in_node.get_meta_info()
    # sub dir
    if in_node.is_dir():
      sub_dir = out_dir.create_dir(name, meta_info, False)
      for child in in_node.get_entries():
        self.repack_node(child, sub_dir)
      sub_dir.flush()
    # file
    elif in_node.is_file():
      data = in_node.get_file_data()
      out_file = out_dir.create_file(name, data, meta_info, False)
      out_file.flush()
    in_node.flush()



from __future__ import absolute_import
from __future__ import print_function

from amitools.fs.block.Block import *
import amitools.fs.DosType as DosType

class PartitionDosEnv:
  valid_keys = ('max_transfer', 'mask', 'num_buffer', 'reserved', 'boot_pri', 'pre_alloc', 'boot_blocks')

  def __init__(self, size=16, block_size=128, sec_org=0, surfaces=0, sec_per_blk=1, blk_per_trk=0,
               reserved=2, pre_alloc=0, interleave=0, low_cyl=0, high_cyl=0, num_buffer=30,
               buf_mem_type=0, max_transfer=0xffffff, mask=0x7ffffffe, boot_pri=0, dos_type=DosType.DOS0,
               baud=0, control=0, boot_blocks=0):
    self.size = size
    self.block_size = block_size
    self.sec_org = sec_org
    self.surfaces = surfaces
    self.sec_per_blk = sec_per_blk
    self.blk_per_trk = blk_per_trk
    self.reserved = reserved
    self.pre_alloc = pre_alloc
    self.interleave = interleave
    self.low_cyl = low_cyl
    self.high_cyl = high_cyl
    self.num_buffer = num_buffer
    self.buf_mem_type = buf_mem_type
    self.max_transfer = max_transfer
    self.mask = mask
    self.boot_pri = boot_pri
    self.dos_type = dos_type
    self.baud = baud
    self.control = control
    self.boot_blocks = boot_blocks

  def dump(self):
    print("DosEnv")
    print(" size:           %d" % self.size)
    print(" block_size:     %d" % self.block_size)
    print(" sec_org:        %d" % self.sec_org)
    print(" surfaces:       %d" % self.surfaces)
    print(" sec_per_blk:    %d" % self.sec_per_blk)
    print(" blk_per_trk:    %d" % self.blk_per_trk)
    print(" reserved:       %d" % self.reserved)
    print(" pre_alloc:      %d" % self.pre_alloc)
    print(" interleave:     %d" % self.interleave)
    print(" low_cyl:        %d" % self.low_cyl)
    print(" high_cyl:       %d" % self.high_cyl)
    print(" num_buffer:     %d" % self.num_buffer)
    print(" buf_mem_type:   0x%08x" % self.buf_mem_type)
    print(" max_transfer:   0x%08x" % self.max_transfer)
    print(" mask:           0x%08x" % self.mask)
    print(" boot_pri:       %d" % self.boot_pri)
    print(" dos_type:       0x%08x = %s" % (self.dos_type, DosType.num_to_tag_str(self.dos_type)))
    print(" baud:           %d" % self.baud)
    print(" control:        %d" % self.control)
    print(" boot_blocks:    %d" % self.boot_blocks)

  def read(self, blk):
    self.size = blk._get_long(32)
    self.block_size = blk._get_long(33)
    self.sec_org = blk._get_long(34)
    self.surfaces = blk._get_long(35)
    self.sec_per_blk = blk._get_long(36)
    self.blk_per_trk = blk._get_long(37)
    self.reserved = blk._get_long(38)
    self.pre_alloc = blk._get_long(39)
    self.interleave = blk._get_long(40)
    self.low_cyl = blk._get_long(41)
    self.high_cyl = blk._get_long(42)
    self.num_buffer = blk._get_long(43)
    self.buf_mem_type = blk._get_long(44)
    self.max_transfer = blk._get_long(45)
    self.mask = blk._get_long(46)
    self.boot_pri = blk._get_slong(47)
    self.dos_type = blk._get_long(48)
    self.baud = blk._get_long(49)
    self.control = blk._get_long(50)
    self.boot_blocks = blk._get_long(51)

  def write(self, blk):
    blk._put_long(32, self.size)
    blk._put_long(33, self.block_size)
    blk._put_long(34, self.sec_org)
    blk._put_long(35, self.surfaces)
    blk._put_long(36, self.sec_per_blk)
    blk._put_long(37, self.blk_per_trk)
    blk._put_long(38, self.reserved)
    blk._put_long(39, self.pre_alloc)
    blk._put_long(40, self.interleave)
    blk._put_long(41, self.low_cyl)
    blk._put_long(42, self.high_cyl)
    blk._put_long(43, self.num_buffer)
    blk._put_long(44, self.buf_mem_type)
    blk._put_long(45, self.max_transfer)
    blk._put_long(46, self.mask)
    blk._put_slong(47, self.boot_pri)
    blk._put_long(48, self.dos_type)
    blk._put_long(49, self.baud)
    blk._put_long(50, self.control)
    blk._put_long(51, self.boot_blocks)

class PartitionBlock(Block):
  FLAG_BOOTABLE = 1
  FLAG_NO_AUTOMOUNT = 2

  def __init__(self, blkdev, blk_num):
    Block.__init__(self, blkdev, blk_num, chk_loc=2, is_type=Block.PART)

  def create(self, drv_name, dos_env, host_id=7, next=Block.no_blk, flags=0, dev_flags=0,
             size=64):
    Block.create(self)
    self.size = size
    self.host_id = host_id
    self.next = next
    self.flags = flags

    self.dev_flags = dev_flags
    self.drv_name = drv_name

    if dos_env == None:
      dos_env = PartitionDosEnv()
    self.dos_env = dos_env
    self.valid = True

  def write(self):
    self._create_data()

    self._put_long(1, self.size)
    self._put_long(3, self.host_id)
    self._put_long(4, self.next)
    self._put_long(5, self.flags)

    self._put_long(8, self.dev_flags)
    self._put_bstr(9, 31, self.drv_name)

    self.dos_env.write(self)

    Block.write(self)

  def read(self):
    Block.read(self)
    if not self.valid:
      return False

    self.size = self._get_long(1)
    self.host_id = self._get_long(3)
    self.next = self._get_long(4)
    self.flags = self._get_long(5)

    self.dev_flags = self._get_long(8)
    self.drv_name = self._get_bstr(9, 31)

    self.dos_env = PartitionDosEnv()
    self.dos_env.read(self)

    return self.valid

  def dump(self):
    Block.dump(self, "Partition")

    print(" size:           %d" % self.size)
    print(" host_id:        %d" % self.host_id)
    print(" next:           %s" % self._dump_ptr(self.next))
    print(" flags:          0x%08x" % self.flags)
    print(" dev_flags:      0x%08x" % self.dev_flags)
    print(" drv_name:       '%s'" % self.drv_name)

    self.dos_env.dump()



# import http.server


DEFAULT_PORT = 25988
DEFAULT_ADDRESS = "127.0.0.1"

# noinspection PyUnresolvedReferences
import fsgs.web.handlers

# class FSGSWebServerHandler(http.server.BaseHTTPRequestHandler):
#
#     # noinspection PyPep8Naming
#     def do_POST(self):
#         pass


class FSGSWebServer(object):
    def run(self):
        # server_class = http.server.HTTPServer
        # handler_class = FSGSWebServerHandler
        # server_address = (DEFAULT_ADDRESS, DEFAULT_PORT)
        print("Listening on {}:{}".format(DEFAULT_ADDRESS, DEFAULT_PORT))
        # httpd = server_class(server_address, handler_class)
        # httpd.serve_forever()

        from bottle import run

        run(host=DEFAULT_ADDRESS, port=DEFAULT_PORT)



import fsui


class TitleBar(fsui.Panel):
    def __init__(self, parent):
        super().__init__(parent, paintable=True)
        self.set_min_height(40)
        self.set_background_color(fsui.Color(0xFF, 0xFF, 0xFF))

        self.menu_button = fsui.Button(self, "=")
        self.minimize_button = fsui.Button(self, "_")
        self.minimize_button.activated.connect(self.on_minimize_button)
        self.maximize_button = fsui.Button(self, "^")
        self.maximize_button.activated.connect(self.on_maximize_button)
        self.close_button = fsui.Button(self, "X")
        self.close_button.activated.connect(self.on_close_button)

        self.layout = fsui.HorizontalLayout()
        self.layout.add(self.menu_button, fill=True)
        self.layout.add_spacer(0, expand=True)
        self.layout.add(self.minimize_button, fill=True)
        self.layout.add(self.maximize_button, fill=True)
        self.layout.add(self.close_button, fill=True)

        self.window_pos = (-1, -1)
        self.mouse_pos = (-1, -1)

    def on_menu_button(self):
        pass

    def on_minimize_button(self):
        self.parent().minimize()

    def on_maximize_button(self):
        self.parent().maximize(not self.parent().is_maximized())

    def on_close_button(self):
        self.parent().close()

    def on_left_dclick(self):
        self.on_maximize_button()

    def on_left_down(self):
        self.window_pos = self.parent().get_position()
        self.mouse_pos = fsui.get_mouse_position()

    def on_mouse_motion(self):
        mouse_pos = fsui.get_mouse_position()
        window_pos = (
            self.window_pos[0] + mouse_pos[0] - self.mouse_pos[0],
            self.window_pos[1] + mouse_pos[1] - self.mouse_pos[1],
        )
        self.parent().set_position(window_pos)

    # def on_left_up(self):
    #     pass

    def on_paint(self):
        dc = self.create_dc()
        if self.close_button is not None:
            x_offset = self.close_button.width() + 20
        else:
            x_offset = 20
        # print(self.parent())
        # text = self.parent().title()
        text = "FIXME"
        tw, th = dc.measure_text(text)
        dc.draw_text(text, x_offset, (self.height() - th) // 2)



from __future__ import absolute_import
from __future__ import print_function

class CommandQueue:
  def __init__(self, cmd_list, sep, cmd_map):
    self.cmd_list = cmd_list
    self.sep = sep
    self.pos = 0
    self.cmd_map = cmd_map
  
  def get_sub_command(self):
    p = self.pos      
    n = len(self.cmd_list)
    if self.pos == n:
      return None
    while p < n:
      if self.cmd_list[p] == self.sep:
        break
      p += 1
    cmd_len = p - self.pos
    cmd = self.cmd_list[self.pos:p]
    if p < n:
      p += 1
    self.pos = p
    if cmd_len == 0:
      return self.get_sub_command()
    else:
      return cmd
  
  def run(self):
    # get first command
    cmd_line = self.get_sub_command()
    if cmd_line == None:
      return 1
    self.cmd_line = cmd_line
    cmd = self.create_cmd_instance(cmd_line)
    if cmd == None:
      self.show_cmd_help(cmd_line[0])
      return 2
    
    # run first command
    exit_code = self.run_first(cmd_line, cmd)
    
    if exit_code == 0:
      cmd_line = self.get_sub_command()
      while cmd_line != None:
        cmd = self.create_cmd_instance(cmd_line)
        if cmd == None:
          self.show_cmd_help(cmd_line[0])
          return 2
  
        # run next command
        exit_code = self.run_next(cmd_line, cmd)
        if exit_code != 0:
          break
        
        cmd_line = self.get_sub_command()
    return exit_code
      
  def run_first(self, cmd_line, cmd):
    return cmd.run()
    
  def run_next(self, cmd_line, cmd):
    return cmd.run()
  
  def create_cmd(self, cclass, name, opts):
    return cclass(name, opts) # implement this!
  
  def create_cmd_instance(self, cmd_line):
    name = cmd_line[0]
    if name in self.cmd_map:
      if len(cmd_line) == 1:
        opts = []
      else:
        opts = cmd_line[1:]
      cclass = self.cmd_map[name]
      return self.create_cmd(cclass, name, opts) 
    else:
      return None
    
  def show_cmd_help(self, name):
    print("INVALID COMMAND:", name)
    print("valid commands are:")
    for a in sorted(self.cmd_map):
      print ("  ", a)
